#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2010, Sebastian Ramacher
# Copyright (C) 2008, Andreas Stuehrk
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

"""
BitTorrent distribution tool client
"""

import os
import re
import sys
import time

from btdt import __version__
from btdt.bencode import bdecode
from btdt.common import AuthorizedServerProxy, size_to_human_readable, mkstemp
from btdt.config import create_client_config_parser
from btdt.logger import init_logger
from btdt.tracker_proxy import TrackerProxy
from threading import Thread

from btdt.torrent_manager import TorrentManager

def encode_for_stdout(string):
    encoding = sys.stdout.encoding or 'ascii'
    return string.encode(encoding, 'replace')


def choose_file(server):
    files = server.list_files()
    if not len(files):
        print 'Keine Dateien verfuegbar!'
        return None

    print 'Verfuegbare Dateien:'
    while True:
        for i, (name, torrents_data) in enumerate(files):
            info = bdecode(torrents_data.data)['info']
            if 'files' in info:
                size = sum((f['length'] for f in info['files']))
            else:
                size = info['length']
            print '  %i) %s (%s)' % (
                i + 1,
                encode_for_stdout(name),
                size_to_human_readable(size)
            )
        try:
            i = raw_input('Datei auswaehlen (a = alle): ')
        except (EOFError, KeyboardInterrupt):
            return None
        if i == 'a':
            return files

        l = i.split(' ')
        valid = True
        for t in l:
            if not t.isdigit():
                valid = False
                break

        if not valid:
            continue

        l = [int(t) - 1 for t in l if int(t) > 0 and int(t) <= len(files)]
        return [files[i] for i in l]

def check_host_from_cmdline(server, port=None, protocol='https'):
    m = re.match('^[a-z]+://[a-zA-Z0-9_.-]+(:\d{1,5})?/?\Z', server)
    if m:
        return server

    # no port?
    if port is not None:
        m = re.match('^[^/]*:\d{1,5}/?\Z', server)
        if not m:
            server = '%s:%d' % (server, port)

    # no protocol?
    m = re.match('^[a-z]+://.*', server)
    if m:
        # something else is wrong
        return None
    else:
        server = '%s://%s' % (protocol, server)

    return check_host_from_cmdline(server)

def main(args=None):
    if args is None:
        args = sys.argv[1:]

    config, parser = create_client_config_parser()
    opts, args = parser.parse_args(args=args)
    if not len(args) == 1:
        parser.print_usage()
        return 1
    init_logger(opts.logfile, opts.loglevel, opts.stdoutloglevel)

    out_dir = args[0]
    host = check_host_from_cmdline(opts.host,
                                   config.getint('Client', 'DefaultRPCPort'),
                                   config.get('Client', 'DefaultRPCProtocol'))
    if host is None:
        print "'%s' is not a valid server!" % (opts.hostm )
        print "Expecting something like [protocol://]hostname[:port]."
        return 1

    username = opts.user
    password = opts.password
    if not username or not password:
        print "'username' or 'password' is empty."
        return 1

    server = AuthorizedServerProxy(host, username, password)
    if opts.mirror:
        print 'downloading all torrents ...'
        files = server.list_files()
    else:
        files = choose_file(server)

    if not files:
        return 1

    tracker_proxy = TrackerProxy(-1, host, username, password)
    proxy_thread = Thread(target=tracker_proxy.serve_forever,
                          name='btdt tracker proxy')
    proxy_thread.start()

    tm = TorrentManager([opts.minport, opts.maxport], maxupload=opts.maxupload,
                       use_natpmp=opts.use_natpmp, use_lsd=opts.use_lsd,
                       use_upnp=opts.use_upnp,
                       fast_resume_path=config.get('libtorrent',
                                                   'FastResumePath'),
                       webseed_workaround=config.getint('libtorrent',
                                                        'WebseedWorkaround'))

    for f, torrent in files:
        # get the torrent ...
        torrent = torrent.data
        f, path = mkstemp()
        os.write(f, torrent)
        os.close(f)

        # ... and add it
        id_ = tm.add_torrent(path, out_dir, opts.storage_mode)
        tm.replace_trackers(
            id_,
            'http://localhost:%i/announce' % tracker_proxy.server_port
        )

    try:
        while True:
            for i in tm.ids():
                state = tm.get_torrent_state(i)
                print '%d: %s (%s, %.2f %%)' % (i,
                                                encode_for_stdout(
                                                    state.name.decode('utf-8')
                                                ),
                                                state.state_string,
                                                state.progress_percent)
                print 'Down/Up: %.2f KiB/s / %.2f KiB/s' % \
                    (state.download_rate_kbs, state.upload_rate_kbs)
                print ''

            time.sleep(opts.update_interval)
    except KeyboardInterrupt:
        pass

    print 'Beende ...'
    tm.quit()
    tracker_proxy.shutdown()
    proxy_thread.join()

    return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
