#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2010, Sebastian Ramacher
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

""" btdt config releated functions
"""

from btdt import __version__
from ConfigParser import ConfigParser
from optparse import OptionParser

import platform
import os
import os.path


def get_config_home():
    """ Return base directory of btdt's configuration files. """
    if 'BTDT_CONFIG_HOME' in os.environ:
        return os.environ.get('BTDT_CONFIG_HOME')

    if platform.system() == 'Windows':
        app_data = os.environ.get('APPDATA')
        return os.path.join(xdg_config_home, 'btdt')
    else:
        xdg_config_home = os.environ.get('XDG_CONFIG_HOME', '~/.config')
        return os.path.expanduser(os.path.join(xdg_config_home, 'btdt'))


# default values
CLIENT_DEFAULT_VALUES = {
    'Client': {
        'Host': '',
        'Minport': 16000,
        'Maxport': 16009,
        'Maxupload': 0,
        'StorageMode': 1,
        'DefaultRPCPort': 15002,
        'DefaultRPCProtocol': 'https',
        # RPC Authorization
        'Username': '',
        'Password': '',
        # GUI
        'Font-Name': 'Lucida Console 10',
        'Background-Color': '#000040',
        'Foreground-Color': '#fff',
        'Download-Color': '#00ff00',
        'Download-Fill': False,
        'Upload-Color': '#ff0000',
        'Upload-Fill': False,

    },
    'ClientLogging': {
        'File': 'None', # disable logging for the client per default
        'Level': 'DEBUG',
        'StdoutLevel': 'WARNING'
    }
}

SERVER_DEFAULT_VALUES = {
    'Tracker': {
        'Host': '',
        'Port': 15001,
        'Torrents_Path': os.path.join(get_config_home(), 'torrents'),
        'Data_Path': 'torrents-data',

        # some more options
        # seconds between expiring downloaders
        'Timeout_Downloaders_Interval': 45 * 60,
        # seconds downloaders should wait between reannouncements
        'Reannounce_Interval': 5 * 60,
        # number of peers to send in an info message
        'Response_Size': 50,
        # how many times to check if a downloader is behind a NAT (0 =
        # don't check)
        'NAT_Check': 3,
        # minutes between reloading of allowed_dir
        'Parse_Allowed_Interval': 15,
        # whether to display names from allowed dir
        'Show_Names': 1,
        # ignore the ip GET parameter from machines which aren't on local
        # network IPs
        'Only_Local_Override_IP': 1,
        # maximum number of peers to give with any one request
        'Max_Give': 200
    },
    'TrackerRPC': {
        'Host': '',
        'Port': 15002
    },
    'Uploader': {
        'Port': 15003,
        'Maxupload': 0,
        'Max_Handler_Threads': 10
    },
    'SSL': {
        'Key': os.path.join(get_config_home(), 'key.pem'),
        'Cert': os.path.join(get_config_home(), 'cert.pem')
    },
    'RemoteAdminLogin': {
        'Enabled': 0,
        'Port': 15004
    },
    'Database': {
        'URL': 'sqlite:///%s' % (os.path.join(get_config_home(),
                                              'database.sqlite'), )
    },
    'Logging': {
        'File': os.path.join(get_config_home(), 'btdt-server.log'),
        'Level': 'DEBUG',
        'StdoutLevel': 'WARNING'
    }
}

LIBTORRENT_DEFAULT_VALUES = {
    'libtorrent': {
        'UseUPnP': 0,
        'UseDHT': 0,
        'UseLSD': 0,
        'UseNatPMP': 0,
        'FastResumePath': os.path.join(get_config_home(), 'fastresume'),
        'WebseedWorkaround': 1
    }
}

def fill_config_default_values(config, default_values):
    for section in default_values.iterkeys():
        if not config.has_section(section):
            config.add_section(section)

        for (opt, val) in default_values[section].iteritems():
            if not config.has_option(section, opt):
                config.set(section, opt, str(val))

def config_section_as_dict(config, section):
    d = {}
    for opt, val in config.items(section):
        d[opt] = val
    return d

def libtorrent_default_config_parser_options(config, parser):
    """
    Adds some libtorrent specific options to the passed optparse.OptionParser
    instance.
    """

    fill_config_default_values(config, LIBTORRENT_DEFAULT_VALUES)

    parser.add_option('--upnp', dest='use_upnp', action='store_true',
                      default=config.getint('libtorrent', 'UseUPnP'))
    parser.add_option('--no-upnp', dest='use_upnp', action='store_false')
    parser.add_option('--dht', dest='use_dht', action='store_true',
                      default=config.getint('libtorrent', 'UseDHT'))
    parser.add_option('--no-dht', dest='use_dht', action='store_false')
    parser.add_option('--nat-pmp', dest='use_natpmp', action='store_true',
                      default=config.getint('libtorrent', 'UseNatPMP'))
    parser.add_option('--no-nat-pmp', dest='use_natpmp', action='store_false')
    parser.add_option('--lsd', dest='use_lsd', action='store_true',
                      default=config.getint('libtorrent', 'UseLSD'))
    parser.add_option('--no-lsd', dest='use_lsd', action='store_false')

def prepare_config_parser(default_values, configparser):
    config = configparser()
    config.read(os.path.join(get_config_home(), 'config.ini'))
    fill_config_default_values(config, default_values)
    return config

def create_client_config_parser(configparser=ConfigParser,
                                optionparser=OptionParser):
    config = prepare_config_parser(CLIENT_DEFAULT_VALUES, configparser)

    usage = 'usage: %prog [options] /out/dir'
    version = '%prog version ' + __version__

    parser = optionparser(usage=usage, version=version)
    parser.add_option('--minport', dest='minport', type='int',
                      default=config.getint('Client', 'Minport'))
    parser.add_option('--maxport', dest='maxport', type='int',
                      default=config.getint('Client', 'Maxport'))
    parser.add_option('--maxupload', dest='maxupload', type='int',
                      default=config.getint('Client', 'Maxupload'),
                      help='maxupload in KiB/s (0 = no limit)')
    parser.add_option('--update-interval', dest='update_interval', type='int',
                      default=2)
    parser.add_option('--storage-mode', dest='storage_mode', type='int',
                      default=config.getint('Client', 'StorageMode'),
                      help='0 = allocate, 1 = sparse, 2 = compact')
    parser.add_option('-m', '--mirror', dest='mirror', action='store_true',
                      default=False)
    parser.add_option('--logfile', dest='logfile', type='string',
                      default=config.get('ClientLogging', 'File'),
                      help='log file')
    parser.add_option('--loglevel', dest='loglevel', type="string",
                      default=config.get('ClientLogging', 'Level'),
                      help='log file log level (one of DEBUG, INFO, WARNING, '
                      'ERROR, CRITICAL)')
    parser.add_option('--stdoutloglevel', dest='stdoutloglevel', type='string',
                      default=config.get('ClientLogging', 'StdoutLevel'),
                      help='log file log level (one of DEBUG, INFO, WARNING, '
                      'ERROR, CRITICAL)')
    parser.add_option('--disable-logging', dest='logfile', action='store_const',
                      const=None, help='disables logging')
    parser.add_option('-H', '--host', dest='host', type='string',
                      default=config.get('Client', 'Host'),
                      help='host to connect to')
    parser.add_option('-u', '--user', dest='user', type='string',
                      default=config.get('Client', 'Username'),
                     help='username')
    parser.add_option('-p', '--password', dest='password', type='string',
                      default=config.get('Client', 'Password'),
                     help='password')
    libtorrent_default_config_parser_options(config, parser)

    return (config, parser)


def create_server_config_parser(configparser=ConfigParser,
                                optionparser=OptionParser):
    config = prepare_config_parser(SERVER_DEFAULT_VALUES, configparser)

    usage = 'usage: %prog [options]'
    version = '%prog version' +  __version__

    parser = optionparser(usage=usage, version=version)
    parser.add_option('--maxupload', dest='maxupload', type='int',
                      default=config.getint('Uploader', 'Maxupload'),
                      help='maxupload in KB/s (0 = no limit)')
    parser.add_option('--logfile', dest='logfile', type='string',
                      default=config.get('Logging', 'File'), help='log file')
    parser.add_option('--loglevel', dest='loglevel', type="string",
                      default=config.get('Logging', 'Level'),
                      help='log file log level (one of DEBUG, INFO, WARNING, '
                      'ERROR, CRITICAL)')
    parser.add_option('--stdoutloglevel', dest='stdoutloglevel', type='string',
                      default=config.get('Logging', 'StdoutLevel'),
                      help='log file log level (one of DEBUG, INFO, WARNING, '
                      'ERROR, CRITICAL)')
    parser.add_option('--disable-logging', dest='logfile', action='store_const',
                      const=None, help='disables logging')

    return (config, parser)
