#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2009, Sebastian Ramacher
# Copyright (C) 2008, Andreas Stuehrk
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

""" BitTorrent tracker

a custom BitTorrent-Tracker based on Bram Cohen's tracker
"""

from __future__ import with_statement

import os
import socket
import time
import btdt.torrent_manager as tm

from btdt.bencode import bencode, Bencached
from btdt.common import path_size, delete_torrent_fastresume, \
                        DefaultingDict, read_torrent

from btdt.database import Torrent, Session
from btdt.logger import logger
from random import shuffle
from threading import Event, Thread

PROTOCOL_NAME = 'BitTorrent protocol'

def nat_check(resultfunc, downloadid, peerid, ip, port):
    to_send = chr(len(PROTOCOL_NAME)) + PROTOCOL_NAME + (chr(0)*8) + downloadid
    to_recv = to_send + peerid

    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((ip, port))
        s.send(to_send)
        data = s.recv(1024)
        s.close()

        if data == to_recv:
            resultfunc(False, downloadid, peerid, ip, port)
            return
    except:
        pass

    resultfunc(True, downloadid, peerid, ip, port)

def parse_torrents(allowed_dir, files_dir, tracker_url, webseed_url):
    """
    Checks file_dir for available downloads, creates torrents in
    allowed_dir if necessary with tracker_url as tracker and webseed_url
    as webseed. Returns a dictionary of available torrents and their data.
    """

    if not os.path.exists(files_dir) or not os.path.isdir(files_dir):
        logger().debug("data directory '%s' doesn't exists" % files_dir)
        return {}
    if not os.path.exists(allowed_dir):
        os.mkdir(allowed_dir)

    for f in os.listdir(files_dir):
        path = os.path.join(files_dir, f)
        logger().debug("found '%s' in data directory" % f)

        torrent_path = os.path.join(allowed_dir, '%s.torrent' % f)
        if not os.path.exists(torrent_path):
            logger().debug("'%s' doesn't exists ... creating it" % torrent_path)
            tm.create_torrent(torrent_path, path, None, webseeds=webseed_url)

    session = Session()
    torrents = {}
    for f in os.listdir(allowed_dir):
        if not f.endswith('.torrent'):
            continue

        path = os.path.join(allowed_dir, f)
        data_path = os.path.join(files_dir, f[:-len('.torrent')])
        if not os.path.exists(data_path):
            # data doesn't exist, delete torrent
            logger().info("'%s''s data doesn't exist ... deleting torrent" % f)
            delete_torrent_fastresume(path)
            continue

        try:
            data, info_hash, info_hash_hex = read_torrent(path)
            info = data['info']
            torrent = {
                'info_hash': info_hash,
                # info['name'] is encoded as utf-8 and f is unicode
                'name': info.get('name', '').decode('utf-8') or f,
                'file': f,
                'path': path,
                'piecelength': info['piece length'],
                'pieces': info['pieces'],
            }

            if 'files' in info:
                torrent['files'] = [
                    os.path.join(torrent['name'],
                                 *[p.decode('utf-8') for p in t['path']])
                    for t in info['files']
                ]
            else:
                torrent['files'] = [torrent['name']]

            l = 0
            if 'length' in info:
                l = info.get('length', 0)
            elif 'files' in info:
                for li in info['files']:
                    if 'length' in li:
                        l = l + li['length']
            torrent['length'] = l

            if path_size(data_path) != l:
                logger().warning("'%s''s meta info doesn't match actual data "
                                 "... deleting it" % path)
                delete_torrent_fastresume(path)

            torrents[info_hash] = torrent

            # add torrent to the database
            q = session.query(Torrent).filter_by(torrent_hash=info_hash_hex)
            if q.count():
                t = q.one()
                t.path = path
                t.size = l
            else:
                session.add(Torrent(info_hash_hex, path, l))
        except Exception, e:
            logger().warning("found invalid torrent '%s' ... deleting it" % f)
            logger().exception(e)
            delete_torrent_fastresume(path)

    logger().debug("valid torrents: %s" % \
                   str([t['path'] for t in torrents.values()]))
    return torrents

def is_valid_ipv4(ip):
    """ Checks whether the given address is a valid IPv4 address. """
    try:
        return compact_peer_info(ip, 0) == 6
    except (ValueError, IndexError):
        return False

def is_local_ip(ip):
    """ Checks whether the given address is a local IP address. """
    try:
        v = [long(x) for x in ip.split('.')]
        if v[0] == 10 or v[0] == 127 or v[:2] in ([192, 168], [169, 254]):
            return True
        if v[0] == 172 and v[1] >= 16 and v[1] <= 31:
            return True
    except ValueError:
        pass

    return False

def compact_peer_info(ip, port):
    return ''.join([chr(int(i)) for i in ip.split('.')]) + \
        chr((port & 0xFF00) >> 8) + chr(port & 0xFF)


class Tracker(object):
    def __init__(self, config, (host, port), webseed_port, key_file, cert_file):
        logger().info('initalizing tracker ...')

        self.server_port = port
        self.doneevent = Event()

        self.tracker_url = 'http://localhost:%d/announce' % port
        self.webseed_url = 'http://%s:%d/webseed/' % (host, webseed_port)

        self.response_size = int(config['response_size'])
        self.natcheck = int(config['nat_check'])
        self.max_give = int(config['max_give'])
        self.reannounce_interval = int(config['reannounce_interval'])
        self.show_names = int(config['show_names'])
        self.only_local_override_ip = int(config['only_local_override_ip'])

        self.becache1 = DefaultingDict({})
        self.becache2 = DefaultingDict({})
        self.cache1 = DefaultingDict([])
        self.cache2 = DefaultingDict([])
        self.times = DefaultingDict({})
        self.state = {
            'peers': DefaultingDict({}),
            'completed': DefaultingDict(0)
        }

        self.downloads = self.state['peers']
        self.completed = self.state['completed']

        self.prevtime = time.time()
        self.timeout_downloaders_interval = \
            int(config['timeout_downloaders_interval'])

        self.allowed = {}
        self.allowed_dir = config['torrents_path'].decode('utf-8')
        self.files_dir = config['data_path'].decode('utf-8')
        self.parse_allowed_interval = int(config['parse_allowed_interval'])
        self.piecesizes = {}

        self.expire_thread = Thread(target=self.expire_downloaders)
        self.expire_thread.start()
        self.parser_thread = Thread(target=self.parse_allowed)
        self.parser_thread.start()

    def shutdown(self):
        logger().info('shutting down tracker ...')

        self.doneevent.set()
        self.expire_thread.join()
        self.parser_thread.join()

        logger().info('tracker shut down done')

    def connectback_result(self, result, downloadid, peerid, ip, port):
        if self.doneevent.isSet():
            return

        logger().debug("got connect-back result for %s (%s:%d)" % \
                       (peerid, ip, port))
        record = self.downloads[downloadid][peerid]
        if record is None or record['ip'] != ip or record['port'] != port:
            return
        if 'nat' not in record:
            record['nat'] = int(not result)
        else:
            if result:
                record['nat'] = 0
            else:
                record['nat'] += 1
        if result:
            data = Bencached(bencode({
                'ip': ip,
                'port': port,
                'peer id': peerid
            }))
            self.becache1[downloadid][peerid] = data
            self.becache2[downloadid][peerid] = compact_peer_info(ip, port)

    def parse_allowed(self):
        while not self.doneevent.isSet():
            logger().info('parsing allowed_dir (interval)')

            self.__parse_allowed()
            self.doneevent.wait(self.parse_allowed_interval * 60)

    def force_parse_allowed(self):
        logger().info('forcing parse_allowed()')

        self.__parse_allowed()

    def __parse_allowed(self):
        new_allowed = parse_torrents(self.allowed_dir, self.files_dir,
                                     self.tracker_url, self.webseed_url)
        self.allowed = new_allowed
        for i in new_allowed.itervalues():
            self.piecesizes[i['name']] = i['piecelength']

    def expire_downloaders(self):
        while not self.doneevent.isSet():
            for x in self.times.keys():
                for myid, t in self.times[x].items():
                    if t < self.prevtime:
                        if myid in self.becache1[x]:
                            del self.becache1[x][myid]
                            del self.becache2[x][myid]
                        del self.times[x][myid]
                        del self.downloads[x][myid]
            self.prevtime = time.time()

            # remove dead items
            for key, value in self.downloads.items():
                if len(value) == 0:
                    del self.times[key]
                    del self.downloads[key]

            self.doneevent.wait(self.timeout_downloaders_interval)

    def delete_torrents(self, unique_ids):
        pass

    def announce(self, client_address, info_hash, params):
        text_no_cache = {'Content-Type': 'text/plain', 'Pragma': 'no-cache'}

        if 'ip' in params and not is_valid_ipv4(params['ip']):
            raise ValueError('DNS name or invalid IP address given for IP')

        infohash = str(info_hash)
        if self.allowed is not None:
            if infohash not in self.allowed:
                return (200, text_no_cache,
                    bencode({'failure reason': 'Requested download is not ' \
                        'authorized for use with this tracker.'}))

        ip = client_address[0]
        ip_override = 0
        if 'ip' in params and \
            (not self.server.only_local_override_ip or is_local_ip(ip)):
            ip_override = 1

        if 'event' in params and \
            params['event'] not in ['started', 'completed', 'stopped']:
            raise ValueError('invalid event')

        port = long(params.get('port', ''))
        uploaded = long(params.get('uploaded', ''))
        downloaded = long(params.get('downloaded', ''))
        left = long(params.get('left', ''))
        myid = params.get('peer_id', '')
        if len(myid) != 20:
            raise ValueError('id not of length 20')

        rsize = self.response_size
        if 'numwant' in params:
            rsize = min(long(params['numwant']), self.max_give)

        peers = self.downloads[infohash]
        ts = self.times[infohash]
        confirm = False
        if myid in peers:
            myinfo = peers[myid]
            if 'key' in myinfo:
                if params['key'] != myinfo['key']:
                    return (200, text_no_cache,
                        bencode({'failure reason': 'key did not match key ' \
                            'supplied earlier'}))

                confirm = True
            elif myinfo['ip'] == ip:
                confirm = True
        else:
            confirm = True

        if params.get('event', '') != 'stopped' and confirm:
            ts[myid] = time.time()
            if myid not in peers:
                peers[myid] = {'ip': ip, 'port': port, 'left': left}
                if 'key' in params:
                    peers[myid]['key'] = params['key']
                if 'ip' in params:
                    peers[myid]['given ip'] = params['ip']

                if ip_override:
                    mip = params['ip']
                else:
                    mip = ip

                if not self.natcheck or ip_override:
                    data = Bencached(bencode({
                        'ip': mip,
                        'port': port,
                        'peer id': myid
                    }))
                    self.becache1[infohash][myid] = data
                    self.becache2[infohash][myid] = compact_peer_info(mip, port)
            else:
                peers[myid]['left'] = left
                peers[myid]['ip'] = ip

            if params.get('event', '') == 'completed':
                self.completed[infohash] += 1

            if port == 0:
                peers[myid]['nat'] = 2**30
            elif self.natcheck and not ip_override:
                to_nat = peers[myid].get('nat', -1)
                if to_nat and to_nat < self.natcheck:
                    Thread(
                        target=nat_check,
                        args=(
                            self.connectback_result,
                            infohash, myid, ip, port)
                    ).start()
            else:
                peers[myid]['nat'] = 0
        elif confirm:
            if myid in peers:
                if myid in self.becache1[infohash]:
                    del self.becache1[infohash][myid]
                    del self.becache2[infohash][myid]
                del peers[myid]
                del ts[myid]

        data = {'interval': self.reannounce_interval}
        if params.get('compact', 0):
            if rsize == 0:
                data['peers'] = ''
            else:
                cache = self.cache2[infohash]
                if len(cache) < rsize:
                    del cache[:]
                    cache.extend(self.becache2[infohash].values())
                    shuffle(cache)
                    del self.cache1[infohash][:]
                data['peers'] = ''.join(cache[-rsize:])
                del cache[-rsize:]
        else:
            if rsize == 0:
                data['peers'] = []
            else:
                cache = self.cache1[infohash]
                if len(cache) < rsize:
                    del cache[:]
                    cache.extend(self.becache1[infohash].values())
                    shuffle(cache)
                    del self.cache2[infohash][:]
                data['peers'] = cache[-rsize:]
                del cache[-rsize:]

        return (200, text_no_cache, bencode(data))

    def scrape(self, info_hash, params):
        names = []
        if info_hash is not None:
            if info_hash in self.downloads:
                names = [info_hash]
        else:
            names = self.downloads.keys()
            names.sort()

        fs = {}
        for name in names:
            l = self.downloads[name]
            n = self.completed.get(name, 0)
            # isinstance(i, dict) should be always true here
            c = len([True for i in l.values() if not i['left']])
            d = len(l) - c
            fs[name] = {'complete': c, 'incomplete': d, 'downloaded': n}
            if self.allowed is not None and name in self.allowed and \
               self.show_names:
                fs[name]['name'] = self.allowed[name]['name']
        r = {'files': fs}

        return (200, {'Content-Type': 'text/plain'}, bencode(r))
