#!/usr/bin/env python
# -*- coding: utf-8 -*-
# $Id$
# Copyright (C) 2008, Sebastian Ramacher
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

""" Database system

Defines classes representing the tables in the database. Also provides a
function to initialize the database and mappers.
"""

from sqlalchemy import create_engine, \
                       Table, Column, Integer, String, MetaData, Boolean, \
                       ForeignKey
from sqlalchemy.orm import mapper, sessionmaker, relation

# check sqlalchemy version
from sqlalchemy import __version__ as sqlalchemy_version
assert sqlalchemy_version >= '0.5', 'btdt requires at least sqlalchemy 0.5'

metadata = MetaData()
Session = sessionmaker(autoflush=True, autocommit=True)

users_table = Table("users", metadata,
    Column("name", String(64), primary_key=True),
    Column("password_hash", String(40)),
    Column("is_admin", Boolean, server_default='FALSE')
)

torrents_table = Table("torrents", metadata,
    Column("torrent_hash", String(40), primary_key=True),
    Column("path", String(256)),
    Column("size", Integer, server_default='0')
)

downloaded_table = Table("downloaded", metadata,
    Column("user_name", String(64), ForeignKey("users.name"), \
           primary_key=True),
    Column("torrent_hash", String(40), ForeignKey("torrents.torrent_hash"), \
           primary_key=True)
)

class User(object):
    def __init__(self, name, password_hash, is_admin=False):
        self.name = name
        self.password_hash = password_hash
        self.is_admin = is_admin

    def __repr__(self):
        return "<users('%s', '%s', '%d')>" % (self.name, self.password_hash,
                                              self.is_admin)

class Torrent(object):
    def __init__(self, torrent_hash, path, size=0):
        self.torrent_hash = torrent_hash
        self.path = path
        self.size = size

    def __repr__(self):
        return "<torrents('%s', '%s', '%d')>" % (self.torrent_hash,
                                                 self.path, self.size)

class Downloaded(object):
    def __init__(self, user_name, torrent_hash):
        self.user_name = user_name
        self.torrent_hash = torrent_hash

    def __repr__(self):
        return "<downloaded('%s', '%s')>" % (self.user_name, self.torrent_hash)

def init_db(db='sqlite:///database.sqlite', echo=False):
    """ Initializes the database system """
    engine = create_engine(db, echo=echo, convert_unicode=True)
    metadata.create_all(engine)

    mapper(User, users_table, properties={
        'downloaded': relation(Downloaded, backref='users')
    })
    mapper(Torrent, torrents_table)
    mapper(Downloaded, downloaded_table, properties={
        'size': relation(Torrent, backref='downloaded')
    })

    Session.configure(bind=engine)
