#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2008, Andreas Stuehrk
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

"""
    guiclient.py
    ~~~~~~~~~~~~

    Client mit PyGTK-GUI für das tolle Torrent-Distributionssystem.
"""

from __future__ import division, with_statement

import atexit
import errno
import os
import socket
import sys
import xmlrpclib
from ConfigParser import ConfigParser
from datetime import timedelta
from math import radians
from operator import itemgetter
from threading import Thread
from time import time
from xmlrpclib import ServerProxy

import cairo
import gobject
import gtk
import pango
gobject.threads_init()

from btdt import torrent_manager as tm, __version__
from btdt.bencode import bdecode
from btdt.common import AuthorizedServerProxy, size_to_human_readable
from btdt.common import mkstemp
from btdt.config import create_client_config_parser
from btdt.logger import init_logger
from btdt.tracker_proxy import TrackerProxy



CONFIG_FILE = 'config.ini'
RUN_FILE = '.running'


# Error-Codes
(
    CONN_REFUSED,
    CONN_TIMED_OUT,
    FORBIDDEN,
    UNKNOWN_PROTOCOL
) = xrange(4)

class Config(ConfigParser, gobject.GObject):
    """Dünner Wrapper um ConfigParser.ConfigParser, der ein
    `changed'-Signal auslöst, wenn eine Option gesetzt wurde."""
    __gsignals__ = {
        'changed': (gobject.SIGNAL_RUN_LAST, None, (str, str, str))
    }
    def __init__(self, *args, **kwargs):
        gobject.GObject.__init__(self)
        ConfigParser.__init__(self, *args, **kwargs)

    def set(self, section, option, value):
        ConfigParser.set(self, section, option, value)
        self.emit('changed', section, option, value)

# XXX HACK?
class TorrentManager(tm.TorrentManager):
    def __init__(self):
        pass

    def init(self, tracker_port, *args, **kwargs):
        tm.TorrentManager.__init__(self, *args, **kwargs)
        self.tracker_port = tracker_port

    def add_torrent(self, *args, **kwargs):
        id_ = tm.TorrentManager.add_torrent(self, *args, **kwargs)
        self.replace_trackers(id_,
                        'http://localhost:%i/announce' % self.tracker_port)
        return id_
# XXX Globales Objekt loswerden
# Globales Manager-Objekt
torrent_manager = TorrentManager()


class Server(AuthorizedServerProxy):
    """Wrapper um `xmlrpclib.ServerProxy'. Ruft bei socket-Exceptions
    automatisch einen Error-Callback auf."""
    class _Method(object):
        def __init__(self, method, error_callback):
            self.__method = method
            self.__error_callback = error_callback

        def __getattr__(self, name):
            method = self.__method.__getattr__(self, name)
            return Server._Method(method, self.__error_callback)

        def __call__(self, *args):
            error_callback = self.__error_callback
            errorcode = None
            try:
                return self.__method(*args)
            except xmlrpclib.ProtocolError, exc:
                if exc.errcode == 403:
                    errorcode = FORBIDDEN
                else:
                    raise
            except socket.timeout:
                errorcode = CONN_TIMED_OUT
            except socket.error, exc:
                error = exc.args[0]
                if error == errno.ECONNREFUSED:
                    errorcode = CONN_REFUSED
            if errorcode is not None and error_callback is not None:
                error_callback(errorcode)

    class _InvalidProtocolMethod(object):
        def __getattr__(self, name):
            return _InvalidProtocolMethod()

        def __call__(self, *args):
            return None

    def __init__(self, host, username, password, error_callback,
                 *args, **kwargs):
        try:
            AuthorizedServerProxy.__init__(self, host, username,
                                           password, *args, **kwargs)
        except IOError:
            self.__invalid_protocol = True
            if error_callback is not None:
                error_callback(UNKNOWN_PROTOCOL)
        else:
            self.__invalid_protocol = False
        self.__error_callback = error_callback

    def __getattr__(self, name):
        if self.__invalid_protocol:
            return Server._InvalidProtocolMethod()
        method = ServerProxy.__getattr__(self, name)
        return Server._Method(method, self.__error_callback)

def get_filelist(host, username, password, callback, error_callback=None):
    server = Server(host, username, password, error_callback)
    result = server.list_files()
    if result is not None:
        callback(result)

def do_preferences(config, parent=None):
    """Einstellungsdialog anzeigen und ggfs Konfiguration
    aktualisieren."""
    dialog = PreferencesDialog(config, parent=parent)
    dialog.run()
    dialog.destroy()

def idle(func):
    """Dekorator für gobject.idle_add"""
    def wrapper(*args):
        gobject.idle_add(func, *args)
    wrapper.__name__ = func.__name__
    return wrapper

class Download(object):
    """Ein Download-Objekt."""
    def __init__(self, host, username, password, torrent_data, outdir,
                 storage_mode, model, iter_, parts_size_getter):
        self.iter = iter_
        self.model = model
        self.row = model[iter_]
        self.row[MainWindow.DOWNLOAD_COLUMN] = self
        self.parts_size_getter = parts_size_getter
        self.torrent_id = None
        self.outdir = outdir
        self.storage_mode = storage_mode

        self.downstat = StatStore(125)
        self.upstat = StatStore(125)

        f, path = mkstemp()
        os.write(f, torrent_data)
        os.close(f)
        self.on_got_torrent(path)

        self.peers = dict()

    def cancel(self, delete=False):
        """Download abbrechen."""
        assert self.torrent_id is not None
        torrent_manager.remove_torrent(self.torrent_id, delete)
        self.torrent_id = None

    def start(self):
        """Download starten."""
        assert self.torrent_id is not None
        torrent_manager.resume_torrent(self.torrent_id)

    def stop(self):
        """Download stoppen."""
        assert self.torrent_id is not None
        torrent_manager.pause_torrent(self.torrent_id)

    @idle
    def on_got_torrent(self, path):
        """Callback. Wird aufgerufen, wenn das Torrent vom Server
        empfangen wurde."""
        self.torrent_id = torrent_manager.add_torrent(
            path,
            self.outdir,
            self.storage_mode
        )
        state = torrent_manager.get_torrent_state(self.torrent_id)
        # XXX TODO
        #self.row[0] = state.name
        self.row[MainWindow.SIZE_COLUMN] = size_to_human_readable(
            state.total_wanted
        )

        gobject.timeout_add(500, self.on_update_status)

    def on_update_status(self):
        """Status aktualisieren.
           Wird alle 500ms automatisch aufgerufen.
        """
        if self.torrent_id is None:
            return False
        state = torrent_manager.get_torrent_state(self.torrent_id)
        dl = size_to_human_readable(state.download_rate) + '/s'
        ul = size_to_human_readable(state.upload_rate) + '/s'
        self.row[MainWindow.STATUS_COLUMN] = state.state_string
        if state.is_paused:
            self.row[MainWindow.STATUS_COLUMN] += ' (gestoppt)'
        self.row[MainWindow.PROGRESS_COLUMN] = state.progress_percent
        self.update_parts(self.iter, state, state.pieces)
        self.row[MainWindow.DOWNRATE_COLUMN] = dl
        self.row[MainWindow.UPRATE_COLUMN] = ul
        # ETA
        delta = timedelta(seconds=state.eta)
        days = delta.days
        hours, seconds = divmod(delta.seconds, 3600)
        minutes, seconds = divmod(seconds, 60)
        if days:
            eta = '%id %ih %im' % (days, hours, minutes)
        elif hours:
            eta = '%ih %im %is' % (hours, minutes, seconds)
        elif minutes:
            eta = '%im %is' % (minutes, seconds)
        elif seconds:
            eta = '%is' % seconds
        else:
            eta = '-'
        self.row[MainWindow.ETA_COLUMN] = eta
        self.downstat.append((time(), state.download_rate_kbs))
        self.upstat.append((time(), state.upload_rate_kbs))

        peers = list(self.peers.iterkeys())
        for peer in torrent_manager.get_peer_info(self.torrent_id) or list():
            name = '%s (%s) ' % (peer.ip, peer.client)
            if name in self.peers:
                # XXX Will raise an exception if there is more than one
                # peer with the same ip
                peers.remove(name)
                iter_ = self.peers[name]
            else:
                iter_ = self.model.append(self.iter)
                self.peers[name] = iter_
                self.model.set(iter_, MainWindow.FILENAME_COLUMN, name)
            dl = size_to_human_readable(peer.down_speed) + '/s'
            ul = size_to_human_readable(peer.up_speed) + '/s'
            self.model.set(iter_, MainWindow.DOWNRATE_COLUMN, ul)
            self.model.set(iter_, MainWindow.PROGRESS_COLUMN,
                           peer.progress * 100)
            self.model.set(iter_, MainWindow.UPRATE_COLUMN, dl)
            self.update_parts(iter_, state, peer.pieces)
        for peer in peers:
            self.model.remove(self.peers[peer])
            del self.peers[peer]

        return True

    def update_parts(self, iter_, state, parts):
        pixbuf = gtk.gdk.Pixbuf(
            gtk.gdk.COLORSPACE_RGB,
            False,
            8,
            *self.parts_size_getter()
        )
        # Fill pixbuf with blue
        pixbuf.fill(0x2020ffff)
        if not state.num_pieces:
            parts = list()
        if parts:
            part_width = pixbuf.get_width() / len(parts)
            if isinstance(parts[0], int):
                orig_parts, parts = parts, list()
                start = -1
                if orig_parts[0]:
                    parts.append([0])
                    start = 0
                for i, (cur, next) in enumerate(zip(orig_parts,
                                                    orig_parts[1:] + [0])):
                    if cur != next:
                        if next:
                            start = i + 1
                            parts.append([start])
                        else:
                            parts[-1].append(i)
            for i, available in enumerate(parts):
                range_first = available[0] * part_width
                range_length = int(
                    (available[1] - available[0] + 1) * part_width
                )
                if range_length:
                    partbuf = pixbuf.subpixbuf(
                        int(range_first), 0,
                        range_length, pixbuf.get_height()
                    )
                    partbuf.fill(0x8080ffff)
        self.model.set(iter_, MainWindow.PARTS_COLUMN, pixbuf)

class ReallyCancelDialog(gtk.MessageDialog):
    def __init__(self, filename, *args, **kwargs):
        #gtk.Dialog.__init__(self,
        #                    flags=(gtk.DIALOG_MODAL |
        #                           gtk.DIALOG_DESTROY_WITH_PARENT),
        #                    buttons=(gtk.STOCK_YES, gtk.RESPONSE_YES,
        #                             gtk.STOCK_NO, gtk.RESPONSE_NO),
        #                    *args, **kwargs)
        text = u'Download von %s wirklich abbrechen?' % filename
        gtk.MessageDialog.__init__(self,
                                   flags=(gtk.DIALOG_MODAL |
                                          gtk.DIALOG_DESTROY_WITH_PARENT),
                                   buttons=gtk.BUTTONS_YES_NO,
                                   message_format=text,
                                   *args, **kwargs)
        self.connect('response', self.on_response)
        self.checkbutton = gtk.CheckButton(u'Datei(en) löschen')
        self.checkbutton.show()
        self.vbox.pack_start(self.checkbutton)

    def on_response(self, dialog, response_id):
        if response_id == gtk.RESPONSE_YES:
            self.delete = dialog.checkbutton.get_active()

class PreferencesDialog(gtk.Dialog):
    """Der Einstellungendialog."""
    def __init__(self, config, title=None, parent=None, *args, **kwargs):
        if title is None:
            title = 'Einstellungen'
        gtk.Dialog.__init__(self,
                            title, parent,
                            flags=(gtk.DIALOG_MODAL |
                                   gtk.DIALOG_DESTROY_WITH_PARENT),
                            buttons=(gtk.STOCK_OK, gtk.RESPONSE_OK,
                                     gtk.STOCK_CANCEL,
                                                gtk.RESPONSE_CANCEL),
                            *args, **kwargs)
        self.config = config
        self.connect('response', self.on_response)
        # Widgets hinzufügen
        notebook = gtk.Notebook()
        self.vbox.pack_start(notebook)
        frame = gtk.Frame()
        notebook.append_page(frame, gtk.Label(u'Allgemein'))
        #
        # "Allgemein"-Seite
        #
        vbox = gtk.VBox()
        frame.add(vbox)
        self.vbox.set_spacing(5)
        # Serveraddresse
        label = gtk.Label('Server:')
        label.set_alignment(0, 0.5)
        vbox.pack_start(label)
        self.host = gtk.Entry()
        vbox.pack_start(self.host)
        # Benutzername
        label = gtk.Label('Benutzername:')
        label.set_alignment(0, 0.5)
        vbox.pack_start(label)
        self.username = gtk.Entry()
        vbox.pack_start(self.username)
        # Passwort
        label = gtk.Label('Passwort:')
        label.set_alignment(0, 0.5)
        vbox.pack_start(label)
        self.password = gtk.Entry()
        self.password.set_invisible_char('*')
        self.password.set_visibility(False)
        vbox.pack_start(self.password)
        # Minport
        label = gtk.Label('Minport:')
        label.set_alignment(0, 0.5)
        vbox.pack_start(label)
        adj = gtk.Adjustment(7001, 0, 2**16 - 1, 1)
        self.minport = gtk.SpinButton(adj)
        vbox.pack_start(self.minport)
        # Maxport
        label = gtk.Label('Maxport:')
        label.set_alignment(0, 0.5)
        vbox.pack_start(label)
        adj = gtk.Adjustment(7009, 0, 2**16 - 1, 1)
        self.maxport = gtk.SpinButton(adj)
        vbox.pack_start(self.maxport)
        # Maxupload
        label = gtk.Label('Uploadbegrenzung (KiB, 0 = unbegrenzt):')
        label.set_alignment(0, 0.5)
        vbox.pack_start(label)
        # 1048576 = 1 GiB
        adj = gtk.Adjustment(0, 0, 1048576, 1)
        self.maxupload = gtk.SpinButton(adj)
        vbox.pack_start(self.maxupload)
        # Outdir
        label = gtk.Label('Outdir:')
        label.set_alignment(0, 0.5)
        vbox.pack_start(label)
        self.outdir = gtk.FileChooserButton('Outdir')
        self.outdir.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
        vbox.pack_start(self.outdir, expand=False)
        vbox.show_all()
        # Werte aus der Config eintragen
        self.host.set_text(config.get('Client', 'Host'))
        self.username.set_text(config.get('Client', 'Username'))
        self.password.set_text(config.get('Client', 'Password'))
        self.maxport.set_value(config.getint('Client', 'Maxport'))
        self.minport.set_value(config.getint('Client', 'Minport'))
        self.maxupload.set_value(config.getint('Client', 'Maxupload'))
        outdir = os.path.abspath(config.get('Client', 'Outdir'))
        self.outdir.set_filename(outdir)
        #
        # "Netgraph"-Seite
        #
        frame = gtk.Frame()
        notebook.append_page(frame, gtk.Label(u'Netgraph'))
        vbox = gtk.VBox()
        vbox.set_spacing(10)
        frame.add(vbox)
        # Global-Frame
        global_frame = gtk.Frame(u'Global')
        vbox.pack_start(global_frame, expand=False)
        global_vbox = gtk.VBox()
        global_frame.add(global_vbox)
        label = gtk.Label(u'Schriftart:')
        label.set_alignment(0, 0.5)
        global_vbox.pack_start(label)
        self.font = gtk.FontButton()
        global_vbox.pack_start(self.font)
        label = gtk.Label(u'Hintergrundfarbe:')
        label.set_alignment(0, 0.5)
        global_vbox.pack_start(label)
        self.bg_color = gtk.ColorButton()
        global_vbox.pack_start(self.bg_color)
        label = gtk.Label(u'Vordergrundfarbe:')
        label.set_alignment(0, 0.5)
        global_vbox.pack_start(label)
        self.fg_color = gtk.ColorButton()
        global_vbox.pack_start(self.fg_color)
        # Download-Frame
        download_frame = gtk.Frame(u'Download')
        vbox.pack_start(download_frame, expand=False)
        download_vbox = gtk.VBox()
        download_frame.add(download_vbox)
        label = gtk.Label(u'Farbe:')
        label.set_alignment(0, 0.5)
        download_vbox.pack_start(label)
        self.download_color = gtk.ColorButton()
        download_vbox.pack_start(self.download_color)
        self.download_fill = gtk.CheckButton(u'Füllen')
        download_vbox.pack_start(self.download_fill)
        # Upload-Frame
        upload_frame = gtk.Frame(u'Upload')
        vbox.pack_start(upload_frame, expand=False)
        upload_vbox = gtk.VBox()
        upload_frame.add(upload_vbox)
        label = gtk.Label(u'Farbe:')
        label.set_alignment(0, 0.5)
        upload_vbox.pack_start(label)
        self.upload_color = gtk.ColorButton()
        upload_vbox.pack_start(self.upload_color)
        self.upload_fill = gtk.CheckButton(u'Füllen')
        upload_vbox.pack_start(self.upload_fill)
        # Werte aus der Config eintragen
        self.font.set_font_name(config.get('Client', 'Font-Name'))
        self.bg_color.set_color(
            gtk.gdk.color_parse(config.get('Client', 'Background-Color'))
        )
        self.fg_color.set_color(
            gtk.gdk.color_parse(config.get('Client', 'Foreground-Color'))
        )
        self.download_color.set_color(
            gtk.gdk.color_parse(config.get('Client', 'Download-Color'))
        )
        self.download_fill.set_active(config.getboolean('Client',
                                                        'Download-Fill'))
        self.upload_color.set_color(
            gtk.gdk.color_parse(config.get('Client', 'Upload-Color'))
        )
        self.upload_fill.set_active(config.getboolean('Client',
                                                      'Upload-Fill'))
        # Anzeigen
        notebook.show_all()

    def on_response(self, dialog, response_id):
        if response_id == gtk.RESPONSE_OK:
            # Allgemein
            host = dialog.host.get_text().strip()
            username = dialog.username.get_text().strip()
            password = dialog.password.get_text().strip()
            maxport = int(dialog.maxport.get_value())
            minport = int(dialog.minport.get_value())
            maxupload = int(dialog.maxupload.get_value())
            outdir = os.path.abspath(dialog.outdir.get_filename())
            # Netgraph
            font = self.font.get_font_name()
            def color_to_string(color):
                colors = ['red', 'green', 'blue']
                string = ''.join('%04x' % getattr(color, name)
                                 for name in colors)
                return '#' + string
            bg_color = color_to_string(dialog.bg_color.get_color())
            fg_color = color_to_string(dialog.fg_color.get_color())
            download_color = color_to_string(
                dialog.download_color.get_color()
            )
            download_fill = dialog.download_fill.get_active()
            upload_color = color_to_string(
                dialog.upload_color.get_color()
            )
            upload_fill = dialog.upload_fill.get_active()
            # Setzen und übernehmen
            # Allgemein
            self.config.set('Client', 'Host', host)
            self.config.set('Client', 'Username', username)
            self.config.set('Client', 'Password', password)
            self.config.set('Client', 'Maxport', str(maxport))
            self.config.set('Client', 'Minport', str(minport))
            self.config.set('Client', 'Maxupload', str(maxupload))
            self.config.set('Client', 'Outdir', outdir)
            torrent_manager.listen(minport, maxport)
            torrent_manager.upload_rate_limit(maxupload)
            # Netgraph
            self.config.set('Client', 'Font-Name', font)
            self.config.set('Client', 'Background-Color', bg_color)
            self.config.set('Client', 'Foreground-Color', fg_color)
            self.config.set('Client', 'Download-Color', download_color)
            self.config.set('Client', 'Download-Fill', str(download_fill))
            self.config.set('Client', 'Upload-Color', upload_color)
            self.config.set('Client', 'Upload-Fill', str(upload_fill))
            # Speichern
            with open(CONFIG_FILE, 'w') as f:
                self.config.write(f)

class StatStore(object):
    """Ein (x,y)-Model für StatGraph. Oder so."""
    def __init__(self, max_entries=50):
        self.x_values = list()
        self.y_values = list()
        self.max_entries = max_entries

    def __iter__(self):
        return iter(zip(self.x_values, self.y_values))

    def __len__(self):
        return len(self.x_values)

    @property
    def max_x_value(self):
        return max(self.x_values)

    @property
    def min_x_value(self):
        return min(self.x_values)

    @property
    def max_y_value(self):
        return max(self.y_values)

    @property
    def min_y_value(self):
        return min(self.y_values)

    def append(self, (x, y)):
        self.x_values.append(x)
        self.y_values.append(y)
        if len(self.x_values) == self.max_entries:
            # XXX eigentlich der kleinste X-Wert?
            self.pop(0)

    def pop(self, index=-1):
        x = self.x_values.pop(index)
        y = self.y_values.pop(index)
        return (x, y)


class StatGraph(gtk.DrawingArea):
    """Grafisches Statistikfoo."""
    (
        MODEL_INDEX,
        COLOR_INDEX,
        FILL_INDEX
    ) = xrange(3)
    def __init__(self, bg_color=(0, 0, 0.25), fg_color=(1, 1, 1),
                 font_name='Terminus 10'):
        gtk.DrawingArea.__init__(self)
        self.bg_color = bg_color
        self.fg_color = fg_color
        self.font_name = font_name
        self.connect('expose-event', self.on_expose)
        self.sources = list()

    def add(self, model, color=None, fill=False):
        if color is None:
            import random
            color = tuple(random.random() for _ in xrange(4))
        self.sources.append((model, color, fill))

    def clear(self):
        self.sources = list()

    def set_bg_color(self, bg_color):
        self.bg_color = bg_color

    def set_fg_color(self, fg_color):
        self.fg_color = fg_color

    def _set_source_value(self, model, index, value):
        models = map(itemgetter(StatGraph.MODEL_INDEX), self.sources)
        model_index = models.index(model)
        row = list(self.sources[model_index])
        row[index] = value
        self.sources[model_index] = tuple(row)

    def set_color(self, model, color):
        self._set_source_value(model, StatGraph.COLOR_INDEX, color)

    def set_fill(self, model, fill):
        self._set_source_value(model, StatGraph.FILL_INDEX, fill)

    def set_font_name(self, font_name):
        self.font_name = font_name

    def get_cairo_context(self, event=None):
        context = self.window.cairo_create()
        if event is not None:
            context.rectangle(
                event.area.x, event.area.y,
                event.area.width, event.area.height
            )
            context.clip()
        return context

    def on_expose(self, widget, event):
        if not self.sources or not any(bool(s[0]) for s in self.sources):
            return
        context = self.get_cairo_context(event)
        self.max_x_value = max(s[0].max_x_value for s in self.sources if
                                                                    s[0])
        self.min_x_value = min(s[0].min_x_value for s in self.sources if
                                                                    s[0])
        self.max_y_value = max(s[0].max_y_value for s in self.sources if
                                                                    s[0])
        self.draw_background(context)
        self.draw_foreground(context)

    def draw_background(self, context):
        """Zeichnet die Achsen, Beschriftung, etcpp."""
        allocation = self.get_allocation()
        context.save()
        layout = context.create_layout()
        layout.set_font_description(pango.FontDescription(self.font_name))
        # Hintergrund füllen
        context.rectangle(0, 0, allocation.width, allocation.height)
        if isinstance(self.bg_color, gtk.gdk.Color):
            context.set_source_color(self.bg_color)
        else:
            context.set_source_rgb(*self.bg_color)
        context.fill()
        # Farbe setzen
        if isinstance(self.fg_color, gtk.gdk.Color):
            context.set_source_color(self.fg_color)
        else:
            context.set_source_rgb(1.0, 1.0, 1.0)
        # Beschriftung X-Achse
        layout.set_text('Zeit (60s)')
        width, height = layout.get_size()
        height /= pango.SCALE
        width /= pango.SCALE
        self.y_max = int(allocation.height - height - 3 - 1)
        context.move_to(allocation.width / 2 - height / 2, self.y_max + 1)
        context.show_layout(layout)
        # Beschriftung Y-Achse
        layout.set_text('%.1f' % self.max_y_value)
        width, height = layout.get_size()
        height /= pango.SCALE
        width /= pango.SCALE
        self.y_min = 5
        self.x_min = int(5 + width + 3 + 1)
        self.x_max = int(allocation.width - 5)
        # 5 px Abstand vom Rand
        context.move_to(5, self.y_min)
        context.show_layout(layout)
        layout.set_text('0')
        width, height = layout.get_size()
        height /= pango.SCALE
        width /= pango.SCALE
        context.move_to(self.x_min - 3 - width, self.y_max - height / 2)
        context.show_layout(layout)
        # KiB/s
        layout.set_text('KiB/s')
        context.rotate(radians(-90))
        context.update_layout(layout)
        width, height = layout.get_size()
        height /= pango.SCALE
        width /= pango.SCALE
        context.move_to(-(allocation.height / 2) - (width / 2) +
                            self.y_min,
                        self.x_min - height - 3)
        context.show_layout(layout)
        context.rotate(radians(90))
        # Rahmen
        context.move_to(self.x_min, self.y_min)
        context.line_to(self.x_max, self.y_min)
        context.line_to(self.x_max, self.y_max)
        context.line_to(self.x_min, self.y_max)
        context.line_to(self.x_min, self.y_min)
        aa = context.get_antialias()
        context.set_antialias(cairo.ANTIALIAS_NONE)
        context.set_line_width(1)
        context.stroke()
        context.set_antialias(aa)
        # Y-Achsen
        context.set_dash([1, 3], 0)
        for fraction in [0.25, 0.5, 0.75]:
            y = int(self.y_min + fraction * (self.y_max - self.y_min))
            context.move_to(self.x_min, y)
            context.line_to(self.x_max, y)
            context.stroke()
        context.restore()

    def draw_foreground(self, context):
        """Zeichnet die Kurven."""
        allocation = self.get_allocation()
        max_value = self.max_y_value
        context.save()
        context.translate(self.x_min, self.y_min)
        context.set_line_width(2)
        for model, color, fill in self.sources:
            if not model:
                continue
            scale_x = (
                allocation.width - self.x_min - (allocation.width -
                                                 self.x_max)
            ) / 60
            if max_value:
                scale_y = (
                    allocation.height - self.y_min - (allocation.height -
                                                      self.y_max)
                ) / max_value
            else:
                scale_y = 0
            if isinstance(color, gtk.gdk.Color):
                context.set_source_color(color)
            else:
                context.set_source_rgba(*color)
            values = iter(sorted(model, key=itemgetter(0), reverse=True))
            # Startpunkt
            x, y = values.next()
            if fill:
                context.move_to(self.x_max - self.x_min,
                                self.y_max - self.y_min)
                context.line_to(self.x_max - self.x_min,
                                (max_value - y) * scale_y)
            else:
                context.move_to((60 - (self.max_x_value - x)) * scale_x,
                                (max_value - y) * scale_y)
            # Kurve zeichnen
            for x, y in values:
                if self.max_x_value - x >= 60:
                    # Letzter Punkt
                    x = self.max_x_value - 60
                    context.line_to(
                        (60 - (self.max_x_value - x)) * scale_x,
                        (max_value - y) * scale_y
                    )
                else:
                    context.line_to(
                        (60 - (self.max_x_value - x)) * scale_x,
                        (max_value - y) * scale_y
                    )
            if fill:
                context.line_to((60 - (self.max_x_value - x)) * scale_x,
                                self.y_max -self.y_min)
                context.fill()
            else:
                context.stroke()
        context.restore()

class MainWindow(gtk.Window):
    """Das Hauptfenster."""
    # Spalten für das Queue-ListStore
    (
        FILENAME_COLUMN,
        SIZE_COLUMN,
        STATUS_COLUMN,
        PROGRESS_COLUMN,
        PARTS_COLUMN,
        DOWNRATE_COLUMN,
        UPRATE_COLUMN,
        ETA_COLUMN,
        DOWNLOAD_COLUMN
    ) = xrange(9)
    # Notebook-Seiten-Reihenfolge
    (
        QUEUE_PAGE,
        FILES_PAGE,
        NETGRAPH_PAGE
    ) = xrange(3)
    def __init__(self, tray, config):
        gtk.Window.__init__(self, gtk.WINDOW_TOPLEVEL)
        self.config = config
        self.connect('delete-event', lambda *args: tray.on_quit(None))
        self.connect('window-state-event', self.on_window_state_event)
        self.is_minimized = False
        self.tray = tray
        self.set_default_size(800, 300)
        self.set_title('Client - %s' % __version__)
        self.server_status = None
        vbox = gtk.VBox()
        # Menü erstellen
        menu = '''
            <ui>
                <popup name="popup_job">
                    <menuitem action="Start" />
                    <menuitem action="Stop" />
                    <menuitem action="Cancel" />
                    <separator />
                    <menuitem action="New Download" />
                    <separator />
                    <menuitem action="Preferences" />
                    <separator />
                    <menuitem action="About" />
                    <menuitem action="Quit" />
                </popup>
                <popup name="popup_nojob">
                    <menuitem action="New Download" />
                    <separator />
                    <menuitem action="Preferences" />
                    <separator />
                    <menuitem action="About" />
                    <menuitem action="Quit" />
                </popup>
                <popup name="popup_filelist">
                    <menuitem action="Start" />
                </popup>
                <toolbar name="toolbar">
                    <toolitem action="Start" />
                    <toolitem action="Stop" />
                    <toolitem action="Cancel" />
                    <separator />
                    <toolitem action="Preferences" />
                    <separator />
                    <toolitem action="About" />
                    <toolitem action="Quit" />
                </toolbar>
            </ui>
        '''
        actions = [
            ('About', gtk.STOCK_ABOUT, u'_Über', None, u'Über Client',
                self.tray.on_about),
            ('Cancel', gtk.STOCK_CANCEL, 'Download _abbrechen',
                None, 'Download entfernen', self.on_cancel_download),
            ('New Download', gtk.STOCK_SAVE, '_Neuer Download',
                None, 'Neuer Download', self.on_new_download),
            ('Preferences', gtk.STOCK_PREFERENCES, '_Einstellungen',
                None, 'Einstellungen', lambda action:
                                            do_preferences(config, self)),
            ('Start', gtk.STOCK_MEDIA_PLAY, 'Download starten',
                None, 'Download starten', self.on_start_stop_download),
            ('Stop', gtk.STOCK_MEDIA_PAUSE, 'Download stoppen',
                None, 'Download stoppen', self.on_start_stop_download),
            (u'Quit', gtk.STOCK_QUIT, u'_Beenden', None,
                u'Client beenden', self.tray.on_quit)
        ]
        ag = gtk.ActionGroup('QueuePopupMenu')
        ag.add_actions(actions)
        manager = gtk.UIManager()
        manager.insert_action_group(ag, 0)
        manager.add_ui_from_string(menu)
        self.menu_job = manager.get_widget('/popup_job')
        self.menu_nojob = manager.get_widget('/popup_nojob')
        self.menu_filelist = manager.get_widget('/popup_filelist')
        # Toolbar
        self.toolbar = manager.get_widget('/toolbar')
        vbox.pack_start(self.toolbar, expand=False)
        # Notebook
        self.notebook = gtk.Notebook()
        self.notebook.connect('switch-page', self.on_notebook_switched)
        vbox.pack_start(self.notebook)
        # Seite "Queue"
        vpaned = gtk.VPaned()
        # Queue-Treeview in einem ScrolledWindow
        sw = gtk.ScrolledWindow()
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        # Dateiname, Größe, Status, Fortschritt, Teile (Pixbuf), Downrate,
        # Uprate, ETA, Download-Objekt
        model = gtk.TreeStore(str, str, str, float, gobject.TYPE_OBJECT,
                              str, str, str, object)
        self.treeview = gtk.TreeView(model)
        self.treeview.connect('button-press-event', self.on_queue_clicked)
        self.treeview.get_selection().set_mode(gtk.SELECTION_SINGLE)
        # Dateiname-Spalte
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(u'Name', renderer,
                                    text=MainWindow.FILENAME_COLUMN)
        column.set_resizable(True)
        column.set_sort_column_id(MainWindow.FILENAME_COLUMN)
        self.treeview.append_column(column)
        # Größen-Spalte
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(u'Größe', renderer,
                                    text=MainWindow.SIZE_COLUMN)
        column.set_resizable(True)
        column.set_sort_column_id(MainWindow.SIZE_COLUMN)
        self.treeview.append_column(column)
        # Status-Spalte
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(u'Status', renderer,
                                    text=MainWindow.STATUS_COLUMN)
        column.set_resizable(True)
        column.set_sort_column_id(MainWindow.STATUS_COLUMN)
        self.treeview.append_column(column)
        # Fortschritt-Spalte
        renderer = gtk.CellRendererProgress()
        column = gtk.TreeViewColumn(u'Fortschritt', renderer,
                                    value=MainWindow.PROGRESS_COLUMN)
        column.set_sort_column_id(MainWindow.PROGRESS_COLUMN)
        column.set_resizable(True)
        self.treeview.append_column(column)
        # Teile-Spalte
        renderer = gtk.CellRendererPixbuf()
        self.parts_column = column = gtk.TreeViewColumn(u'Teile', renderer,
                                    pixbuf=MainWindow.PARTS_COLUMN)
        column.set_resizable(True)
        self.treeview.append_column(column)
        # DL-Rate-Spalte
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(u'DL-Rate', renderer,
                                    text=MainWindow.DOWNRATE_COLUMN)
        column.set_resizable(True)
        column.set_sort_column_id(MainWindow.DOWNRATE_COLUMN)
        self.treeview.append_column(column)
        # UL-Rate-Spalte
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(u'UL-Rate', renderer,
                                    text=MainWindow.UPRATE_COLUMN)
        column.set_resizable(True)
        column.set_sort_column_id(MainWindow.UPRATE_COLUMN)
        self.treeview.append_column(column)
        # ETA-Spalte
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(u'Restzeit', renderer,
                                    text=MainWindow.ETA_COLUMN)
        column.set_resizable(True)
        column.set_sort_column_id(MainWindow.ETA_COLUMN)
        self.treeview.append_column(column)
        self.treeview.columns_autosize()
        sw.add(self.treeview)
        vpaned.pack1(sw)
        # Netgraph für eine bestimmte Datei
        bg_color = gtk.gdk.color_parse(
            config.get('Client', 'Background-Color')
        )
        fg_color = gtk.gdk.color_parse(
            config.get('Client', 'Foreground-Color')
        )
        font = config.get('Client', 'Font-Name')
        self.treeview.get_selection().connect('changed',
                                        self.on_queue_selection_changed)
        self.file_netgraph = StatGraph(
            bg_color=bg_color,
            fg_color=fg_color,
            font_name=font
        )
        # Alle 5 Sekunden aktualisieren
        gobject.timeout_add(500, lambda:
                                    not self.file_netgraph.queue_draw())
        vpaned.pack2(self.file_netgraph, resize=False)
        self.notebook.append_page(vpaned, gtk.Label('Queue'))
        # Seite "Verfügbare Dateien"
        sw = gtk.ScrolledWindow()
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        # Name als UTF-8-String, torrent_data, Größe als str
        model = gtk.ListStore(str, object, str)
        self.available_treeview = gtk.TreeView(model)
        self.available_treeview.connect('button-press-event',
                                        self.on_filelist_clicked)
        self.parts_renderer = renderer = gtk.CellRendererPixbuf()
        renderer.set_property('stock-id', gtk.STOCK_SAVE)
        column = gtk.TreeViewColumn(u'Ziehen', renderer)
        self.available_treeview.append_column(column)
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(u'Name', renderer, text=0)
        column.set_sort_column_id(0)
        self.available_treeview.append_column(column)
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(u'Größe', renderer, text=2)
        self.available_treeview.append_column(column)
        column.set_sort_column_id(2)
        sw.add(self.available_treeview)
        self.notebook.append_page(sw,
                                  gtk.Label(u'Verfügbare Dateien'))
        # "Netgraph"-Seite
        self.netgraph = StatGraph(
            bg_color=bg_color,
            fg_color=fg_color,
            font_name=font
        )
        self.downstat = StatStore(125)
        self.upstat = StatStore(125)
        download_color = gtk.gdk.color_parse(
            config.get('Client', 'Download-Color')
        )
        self.netgraph.add(
            self.downstat,
            download_color,
            config.getboolean('Client', 'Download-Fill')
        )
        upload_color = gtk.gdk.color_parse(
            config.get('Client', 'Upload-Color')
        )
        self.netgraph.add(
            self.upstat,
            upload_color,
            config.getboolean('Client', 'Upload-Fill')
        )
        gobject.timeout_add(500, self.on_update_netgraph, self.netgraph)
        self.notebook.append_page(self.netgraph, gtk.Label(u'Netgraph'))
        # Status-Bar
        self.statusbar = gtk.Statusbar()
        vbox.pack_start(self.statusbar, expand=False)
        vbox.show_all()
        self.add(vbox)
        # Timer für die Statusbar
        gobject.timeout_add(500, self.on_update_statusbar)
        # Änderungen in der Konfiguration sollen sofort übernommen
        # werden
        config.connect('changed', self.on_config_changed)

    def on_cancel_download(self, action):
        selection = self.treeview.get_selection()
        model, iter_ = selection.get_selected()
        if iter_ is None:
            # Waaah. Keine Reihe ausgewählt. Sollte eigtl nicht
            # funktionieren.
            return
        download = model.get_value(iter_, MainWindow.DOWNLOAD_COLUMN)
        filename = model.get_value(iter_, MainWindow.FILENAME_COLUMN)
        dialog = ReallyCancelDialog(filename=filename, parent=self)
        if dialog.run() == gtk.RESPONSE_YES:
            download.cancel(dialog.delete)
            model.remove(iter_)
        dialog.destroy()

    def on_new_download(self, action):
        self.notebook.set_current_page(MainWindow.FILES_PAGE)

    def on_start_stop_download(self, action):
        if self.notebook.get_current_page() == MainWindow.QUEUE_PAGE:
            # Aktuelle Seite des Notebooks ist die Queue
            selection = self.treeview.get_selection()
            model, iter_ = selection.get_selected()
            if iter_ is None:
                return
            download = model.get_value(iter_, MainWindow.DOWNLOAD_COLUMN)
            if action.get_name() == 'Start':
                download.start()
            else:
                download.stop()
        else:
            # Aktuelle Seite des Notebooks sind die verfügbaren Dateien
            self.add_download()

    def on_notebook_switched(self, notebook, page, page_num):
        sensitive = False
        if page_num == MainWindow.FILES_PAGE:
            self.available_treeview.get_model().clear()
            Thread(
                target=get_filelist,
                args=(
                    self.config.get('Client', 'Host'),
                    self.config.get('Client', 'Username'),
                    self.config.get('Client', 'Password'),
                    self.on_got_filelist,
                    self.on_server_error
                )
            ).start()
        elif page_num == MainWindow.QUEUE_PAGE:
            sensitive = True
        # Die ersten drei Buttons in der Toolbar (de)aktivieren
        for i in xrange(3):
            item = self.toolbar.get_nth_item(i)
            item.set_sensitive(sensitive)

    @idle
    def on_got_filelist(self, files):
        """Wird aufgerufen, wenn die Dateiliste vom Server empfangen
        wurde."""
        self.server_status = None
        # Liste mit Dateien, die bereits gezogen werden
        queue = map(itemgetter(MainWindow.FILENAME_COLUMN),
                    self.treeview.get_model())
        model = self.available_treeview.get_model()
        model.clear()
        for name, torrent_data in files:
            name = name.encode('utf-8')
            if not name in queue:
                torrent_data = torrent_data.data
                info = bdecode(torrent_data)['info']
                if 'files' in info:
                    size = sum((f['length'] for f in info['files']))
                else:
                    size = info['length']
                size = size_to_human_readable(size)
                model.append([name, torrent_data, size])

    def on_filelist_clicked(self, treeview, event):
        path = treeview.get_path_at_pos(int(event.x), int(event.y))
        if path is None:
            return
        if event.type == gtk.gdk._2BUTTON_PRESS:
            self.add_download()
        elif event.type == gtk.gdk.BUTTON_PRESS:
            if event.button == 1:
                column = path[1]
                # XXX Schönerer Weg?
                renders = column.get_cell_renderers()
                if renders and isinstance(renders[0],
                                          gtk.CellRendererPixbuf):
                    selection = self.available_treeview.get_selection()
                    selection.select_path(path[0])
                    self.add_download()
            elif event.button == 3:
                self.menu_filelist.popup(None, None, None, event.button,
                                         event.time)

    def add_download(self, filename=None):
        if filename is None:
            selection = self.available_treeview.get_selection()
            model, iter_ = selection.get_selected()
            if iter_ is None:
                # Sollte niemals passieren
                return
            filename = model.get_value(iter_, 0)
            torrent_data = model.get_value(iter_, 1)
            model.remove(iter_)
        else:
            raise NotImplementedError()
        model = self.treeview.get_model()
        from operator import sub
        getter = lambda: map(
            sub,
            tuple(
                self.treeview.get_cell_area(
                    self.treeview.get_visible_range()[0],
                    self.parts_column
                )
            )[2:],
            [2, 2]
        )
        pixbuf = gtk.gdk.Pixbuf(
            gtk.gdk.COLORSPACE_RGB,
            False,
            8,
            1, 1
        )
        pixbuf.fill(0x0000ffff)
        iter_ = model.append(None, [filename, '-', '', 0, pixbuf, '-',
                             '-', '-', None])
        Download(
            self.config.get('Client', 'Host'),
            self.config.get('Client', 'Username'),
            self.config.get('Client', 'Password'),
            torrent_data,
            self.config.get('Client', 'Outdir'),
            self.config.getint('Client', 'StorageMode'),
            model, iter_,
            getter
        )

    def on_queue_clicked(self, treeview, event):
        if event.type == gtk.gdk.BUTTON_PRESS and event.button == 3:
            path = treeview.get_path_at_pos(int(event.x), int(event.y))
            if path is not None:
                self.menu_job.popup(None, None, None, event.button,
                                    event.time)
            else:
                self.menu_nojob.popup(None, None, None, event.button,
                                      event.time)

    def on_queue_selection_changed(self, selection):
        model, iter_ = selection.get_selected()
        if iter_ is None:
            return
        iter_ = self.treeview.get_model().iter_parent(iter_) or iter_
        if not self.file_netgraph.sources:
            # Größe setzen
            self.file_netgraph.set_size_request(
                self.file_netgraph.get_allocation().width,
                int(self.get_allocation().height / 4)
            )
        # Modelle der Netgraph-Anzeige aktualisieren
        download = model.get_value(iter_, MainWindow.DOWNLOAD_COLUMN)
        self.file_netgraph.clear()
        download_color = gtk.gdk.color_parse(
            self.config.get('Client', 'Download-Color')
        )
        self.file_netgraph.add(
            download.downstat,
            download_color,
            self.config.getboolean('Client', 'Download-Fill')
        )
        upload_color = gtk.gdk.color_parse(
            self.config.get('Client', 'Upload-Color')
        )
        self.file_netgraph.add(
            download.upstat,
            upload_color,
            self.config.getboolean('Client', 'Upload-Fill')
        )

    def on_update_netgraph(self, netgraph):
        """Wird alle 500ms aufgerufen, um den gesamt-netgraph zu
        aktualisieren."""
        if self.notebook.get_current_page() == MainWindow.NETGRAPH_PAGE:
            netgraph.queue_draw()
        return True

    def on_update_statusbar(self):
        """Wird alle 500ms aufgerufen, um die Statusbar zu
        aktualisieren."""
        state = torrent_manager.get_state()
        self.statusbar.pop(0)
        downloaded = size_to_human_readable(state.total_downloaded)
        uploaded = size_to_human_readable(state.total_uploaded)
        status_text = (
            'D: %i KiB/s (gesamt: %s)  '
            'U: %i KiB/s (gesamt: %s)  '
            'Server: %s ' % (state.download_rate_kbs,
                             downloaded,
                             state.upload_rate_kbs,
                             uploaded,
                             self.config.get('Client', 'Host'))
        )
        if self.server_status is not None:
            status_text += u' (%s)' % self.server_status
        self.statusbar.push(0, status_text)
        self.downstat.append([time(), state.download_rate_kbs])
        self.upstat.append([time(), state.upload_rate_kbs])
        return True

    def on_window_state_event(self, window, event):
        if event.changed_mask & gtk.gdk.WINDOW_STATE_ICONIFIED:
            if (event.new_window_state & gtk.gdk.WINDOW_STATE_ICONIFIED
                and not self.is_minimized):
                # Minimiert
                self.set_skip_taskbar_hint(True)
                self.is_minimized = True
            else:
                # Nicht minimiert
                self.present()
                self.set_skip_taskbar_hint(False)
                self.is_minimized = False

    def on_config_changed(self, config, section, option, value):
        """Wird bei Änderungen in der Konfiguration aufgerufen. Setzt
        die Änderungen in der GUI um."""
        if section != 'Client':
            return
        option = option.lower()
        if option == 'background-color':
            bg_color = gtk.gdk.color_parse(value)
            self.file_netgraph.set_bg_color(bg_color)
            self.netgraph.set_bg_color(bg_color)
        elif option == 'foreground-color':
            fg_color = gtk.gdk.color_parse(value)
            self.file_netgraph.set_fg_color(fg_color)
            self.netgraph.set_fg_color(fg_color)
        elif option in ['download-color', 'download-fill',
                        'upload-color', 'upload-fill']:
            rate_type, option = option.split('load-')
            if option == 'color':
                value = gtk.gdk.color_parse(value)
            else:
                value = value.lower() == 'true'
            getattr(self.netgraph, 'set_' + option)(
                getattr(self, rate_type + 'stat'),
                value
            )
            model, iter_ = self.treeview.get_selection().get_selected()
            if iter_ is not None:
                download = model.get_value(iter_,
                                           MainWindow.DOWNLOAD_COLUMN)
                getattr(self.file_netgraph, 'set_' + option)(
                    getattr(download, rate_type + 'stat'),
                    value
                )
        elif option == 'font-name':
            self.file_netgraph.set_font_name(value)
            self.netgraph.set_font_name(value)
        elif option == 'host':
            self.server_status = None

    @idle
    def on_server_error(self, errorcode):
        err_to_string = {
            CONN_REFUSED: 'Verbindung abgelehnt',
            CONN_TIMED_OUT: 'Server antwortet nicht',
            FORBIDDEN: u'Benutzername oder Passwort ungültig',
            UNKNOWN_PROTOCOL: u'Ungültiges XML-RPC-Protokoll'
        }
        self.server_status = err_to_string[errorcode]
        dialog = gtk.MessageDialog(
            parent=self,
            flags=gtk.DIALOG_MODAL,
            buttons=gtk.BUTTONS_OK,
            type=gtk.MESSAGE_ERROR,
            message_format=(u'Der Server scheint nicht erreichbar '
                            u'zu sein: %s.' % err_to_string[errorcode])
        )
        dialog.set_title(u'Server nicht erreichbar')
        dialog.run()
        dialog.destroy()

class TrayIcon(gtk.StatusIcon):
    def __init__(self, config):
        gtk.StatusIcon.__init__(self)
        self.config = config
        self.window = MainWindow(self, config)
        # Popup-Menü erstellen
        menu = '''
            <ui>
                <popup>
                    <menuitem action="Preferences" />
                    <separator />
                    <menuitem action="About" />
                    <menuitem action="Quit" />
                </popup>
            </ui>
        '''
        actions = [
            ('About', gtk.STOCK_ABOUT, u'_Über', None,
                u'Über Client', self.on_about),
            ('Preferences', gtk.STOCK_PREFERENCES, '_Einstellungen',
                None, 'Einstellungen', lambda action:
                                    do_preferences(config, self.window)),
            ('Quit', gtk.STOCK_QUIT, '_Beenden', None,
                'Client beenden', self.on_quit)
        ]
        ag = gtk.ActionGroup('TrayPopupMenu')
        ag.add_actions(actions)
        manager = gtk.UIManager()
        manager.insert_action_group(ag, 0)
        manager.add_ui_from_string(menu)
        self.menu = manager.get_widget('/popup')
        # Icon setzen
        path = os.path.abspath(os.path.dirname(__file__))
        iconfile = os.path.join(path, 'icons', 'tray.png')
        self.set_from_file(iconfile)
        self.set_visible(True)
        # Queue-Fenster-Icon setzen
        self.window.set_icon(self.get_pixbuf())
        # Signale
        self.connect('activate', self.on_activate)
        self.connect('popup-menu', self.on_popup_menu)
        # Refresh tooltip all 5 seconds
        gobject.timeout_add(5000, self.on_update_tooltip)

    def on_about(self, action):
        """Zeigt den About-Dialog an."""
        # Über-Dialog bauen
        dialog = gtk.AboutDialog()
        dialog.set_name('Client')
        dialog.set_version(__version__)
        dialog.set_comments(u'Ein grafischer Client für das tolle '
                            u'Torrent-Distributionssystem.')
        dialog.set_license(u'Total proprietär.\nCopyright 2008 bei mir.')
        dialog.set_authors((u'Ich',))
        dialog.set_logo(self.get_pixbuf())
        dialog.run()
        dialog.destroy()

    def on_quit(self, action):
        text = u'Client wirklich beenden?'
        dialog = gtk.MessageDialog(parent=self.window,
                                   flags=(gtk.DIALOG_MODAL |
                                          gtk.DIALOG_DESTROY_WITH_PARENT),
                                   type=gtk.MESSAGE_QUESTION,
                                   buttons=gtk.BUTTONS_YES_NO,
                                   message_format=text)
        if dialog.run() == gtk.RESPONSE_YES:
            self.set_visible(False)
            # Hide all windows first
            dialog.destroy()
            self.window.hide()
            gtk.main_quit()
        else:
            dialog.destroy()
            return True

    def on_activate(self, action):
        self.window.present()

    def on_popup_menu(self, status, button, time):
        self.menu.popup(None, None, None, button, time)

    def on_update_tooltip(self):
        state = torrent_manager.get_state()
        dl = size_to_human_readable(state.download_rate) + '/s'
        ul = size_to_human_readable(state.upload_rate) + '/s'
        self.set_tooltip(
            'Aktive Downloads: %i (Download: %s, '
            'Upload: %s)' % (len(torrent_manager.torrents), dl, ul)
        )
        return True

def main(args=None):
    """Main-Funktion."""
    if args is None:
        args = sys.argv[1:]

    config, parser = create_client_config_parser(configparser=Config)
    opts, args = parser.parse_args(args=args)
    if args:
        parser.print_usage()
        return 1
    else:
        config.set('Client', 'Host', opts.host)
    init_logger(opts.logfile, opts.loglevel, opts.stdoutloglevel)

    # XXX Code aufräumen
    if os.path.exists(RUN_FILE):
        from threading import Event
        start_client = Event()
        port = -1
        with open(RUN_FILE, 'r') as f:
            port = int(f.read())
        if port != -1:
            try:
                ServerProxy('http://localhost:%i/' % port).present()
            except socket.error:
                # Der Client scheint tot zu sein
                dialog = gtk.MessageDialog(
                    flags=gtk.DIALOG_MODAL,
                    type=gtk.MESSAGE_QUESTION,
                    buttons=gtk.BUTTONS_YES_NO,
                    message_format=(u'Es scheint bereits eine Instanz '
                                    u'zu laufen, doch diese reagiert '
                                    u'nicht. Lockfile löschen und '
                                    u'Client starten?')
                )
                dialog.set_title(u'Client läuft bereits')
                if dialog.run() == gtk.RESPONSE_YES:
                    os.unlink(RUN_FILE)
                    start_client.set()
                dialog.destroy()
        if start_client.isSet():
            return main()
        else:
            return 1
    else:
        tray = TrayIcon(config)

        from SimpleXMLRPCServer import SimpleXMLRPCServer
        server = SimpleXMLRPCServer(('localhost', socket.INADDR_ANY))
        def present():
            gobject.idle_add(tray.window.present)
            return True
        server.register_function(present)
        with open(RUN_FILE, 'w') as f:
            f.write(str(server.socket.getsockname()[1]))
        t = Thread(target=server.serve_forever)
        t.setDaemon(True)
        t.start()
        atexit.register(os.unlink, RUN_FILE)

        tracker_proxy = TrackerProxy(
            -1,
            config.get('Client', 'Host'),
            config.get('Client', 'Username'),
            config.get('Client', 'Password')
        )
        tracker_proxy_port = tracker_proxy.server_port
        def change_tracker_proxy_host(config, section, option, value):
            # XXX irgendwie ein hack..
            option = option.lower()
            if section.lower() != 'client':
                return
            if option in ['host', 'password', 'username']:
                username = config.get('Client', 'Username')
                password = config.get('Client', 'Password')
                tracker_proxy.rpc = AuthorizedServerProxy(
                    config.get('Client', 'Host'),
                    username, password
                )
        config.connect('changed', change_tracker_proxy_host)
        proxy_thread = Thread(target=tracker_proxy.serve_forever,
                              name='btdt tracker proxy')
        proxy_thread.start()

        torrent_manager.init(
            tracker_proxy_port, [opts.minport, opts.maxport],
            maxupload=opts.maxupload,
            fast_resume_path=config.get('libtorrent', 'FastResumePath'),
            webseed_workaround=config.getint('libtorrent', 'WebseedWorkaround')
        )

        tray.set_visible(True)
        tray.window.present()
        try:
            gtk.main()
        except KeyboardInterrupt:
            pass
        finally:
            torrent_manager.quit()
            tracker_proxy.shutdown()
            proxy_thread.join()
        return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
