# -*- coding: utf-8 -*-

# Taken from http://download.bittorrent.com/dl/BitTorrent-5.3-GPL.tar.gz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Written by Bram Cohen
# Modifications for use in Deluge by Andrew Resch 2008
# Modifications for use in btdt by Sebastian Ramacher 2009-2010

from __future__ import with_statement

import os
import sys
import time
from hashlib import sha1 as sha

from btdt.bencode import bencode
from btdt.logger import logger as log

ignore = ['core', 'CVS', 'Thumbs.db', 'desktop.ini', '.svn', '.git', '.hg']

noncharacter_translate = {}
for i in xrange(0xD800, 0xE000):
    noncharacter_translate[i] = ord('-')
for i in xrange(0xFDD0, 0xFDF0):
    noncharacter_translate[i] = ord('-')
for i in (0xFFFE, 0xFFFF):
    noncharacter_translate[i] = ord('-')

def gmtime():
    return time.mktime(time.gmtime())

def get_filesystem_encoding():
    default_encoding = 'utf8'

    if os.path.supports_unicode_filenames:
        encoding = None
    else:
        try:
            encoding = sys.getfilesystemencoding()
        except AttributeError:
            log().debug("This version of Python cannot detect filesystem "
                        "encoding.")

        if encoding is None:
            encoding = default_encoding
            log().debug("Python failed to detect filesystem encoding. "
                      "Assuming '%s' instead.", default_encoding)
        else:
            try:
                'a1'.decode(encoding)
            except:
                log().debug("Filesystem encoding '%s' is not supported. "
                            "Using '%s' instead.", encoding, default_encoding)
                encoding = default_encoding

    return encoding

def decode_from_filesystem(path):
    encoding = get_filesystem_encoding()
    if encoding is None:
        return path
    else:
        assert isinstance(path, str), "Path should be str not %s" % type(path)
        return path.decode(encoding)

def make_meta_file(path, url, piece_length, title=None, comment=None,
                   safe=None, target=None, url_list=None, name=None,
                   private=False, created_by=None, httpseeds=None):
    # piece_length is in KiB!
    piece_length *= 1024

    data = {'creation date': int(gmtime())}
    if url:
        data['announce'] = url.strip()
    a, b = os.path.split(path)
    if not target:
        if b == '':
            f = a + '.torrent'
        else:
            f = os.path.join(a, b + '.torrent')
    else:
        f = target
    info = makeinfo(path, piece_length, name, private)

    data['info'] = info
    if title:
        data['title'] = title
    if comment:
        data['comment'] = comment
    if safe:
        data['safe'] = safe
    if url_list:
        data['url-list'] = url_list
    if created_by:
        data['created by'] = created_by
    if httpseeds:
        data['httpseeds'] = httpseeds

    with open(f, 'wb') as h:
        h.write(bencode(data))

def calcsize(path):
    total = 0
    for s in subfiles(os.path.abspath(path)):
        total += os.path.getsize(s[1])
    return total

def to_utf8(name):
    if isinstance(name, unicode):
        u = name
    else:
        try:
            u = decode_from_filesystem(name)
        except Exception, e:
            raise Exception('Could not convert file/directory name %r to '
                            'Unicode. Either the assumed filesystem '
                            'encoding "%s" is wrong or the filename contains '
                            'illegal bytes.' % (name,
                                                get_filesystem_encoding()))

    if u.translate(noncharacter_translate) != u:
        raise Exception('File/directory name "%s" contains reserved '
                        'unicode values that do not correspond to '
                        'characters.' % name)
    return u.encode('utf-8')

def makeinfo(path, piece_length, name = None, private=False):
    path = os.path.abspath(path)
    piece_count = 0
    if os.path.isdir(path):
        subs = subfiles(path)
        subs.sort()
        pieces = []
        sh = sha()
        done = 0
        fs = []
        totalsize = 0.0
        totalhashed = 0
        for p, f in subs:
            totalsize += os.path.getsize(f)
        if totalsize >= piece_length:
            num_pieces = totalsize / piece_length
        else:
            num_pieces = 1

        for p, f in subs:
            pos = 0
            size = os.path.getsize(f)
            p2 = [to_utf8(n) for n in p]
            fs.append({'length': size, 'path': p2})
            with open(f, 'rb') as h:
                while pos < size:
                    a = min(size - pos, piece_length - done)
                    sh.update(h.read(a))
                    piece_count += 1
                    done += a
                    pos += a
                    totalhashed += a

                    if done == piece_length:
                        pieces.append(sh.digest())
                        done = 0
                        sh = sha()
        if done > 0:
            pieces.append(sh.digest())

        if name is not None:
            assert isinstance(name, unicode)
            name = to_utf8(name)
        else:
            name = to_utf8(os.path.split(path)[1])

        return {'pieces': ''.join(pieces),
            'piece length': piece_length, 'files': fs,
            'name': name,
            'private': private}
    else:
        size = os.path.getsize(path)
        if size >= piece_length:
            num_pieces = size / piece_length
        else:
            num_pieces = 1

        pieces = []
        p = 0
        with open(path, 'rb') as h:
            while p < size:
                x = h.read(piece_length)
                pieces.append(sha(x).digest())
                piece_count += 1
                p += piece_length
                if p > size:
                    p = size
        return {'pieces': ''.join(pieces),
            'piece length': piece_length, 'length': size,
            'name': to_utf8(os.path.split(path)[1]),
            'private': private}

def subfiles(d):
    r = []
    stack = [([], d)]
    while stack:
        p, n = stack.pop()
        if os.path.isdir(n):
            for s in os.listdir(n):
                if s not in ignore and not s.startswith('.'):
                    stack.append((p + [s], os.path.join(n, s)))
        else:
            r.append((p, n))
    return r
