#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2010, Sebastian Ramacher
# Copyright (C) 2008, Andreas Stuehrk
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

""" BitTorrent distribution tool server

Provides the server's main function.
"""

import os.path
import sys

from btdt.admin_rpc import AdminRPC
from btdt.config import create_server_config_parser, config_section_as_dict
from btdt.custom_tracker import Tracker
from btdt.database import init_db
from btdt.logger import init_logger, logger
from btdt.secure_server import create_certificate
from btdt.tracker_rpc import TrackerRPC
from btdt.webseeder import Webseeder
from threading import Thread


def main(args=None):
    """ Runs the server. """
    if args is None:
        args = sys.argv[1:]

    config, parser = create_server_config_parser()
    opts, args = parser.parse_args(args=args)
    if len(args):
        parser.print_usage()
        return 1

    init_logger(opts.logfile, opts.loglevel, opts.stdoutloglevel)
    log = logger()
    log.info('===== starting log =====')

    # initialize server
    log.info('initializing btdt-server ...')

    # check existence of key and certificate
    key_file = config.get('SSL', 'Key')
    cert_file = config.get('SSL', 'Cert')
    if not os.path.exists(key_file) or not os.path.exists(cert_file):
        log.info('creating key and certificate ...')
        create_certificate(config.get('Tracker', 'Host'), key_file, cert_file)

    webseed_port = config.getint('Uploader', 'Port')
    rpc_host = config.get('TrackerRPC', 'Host')
    rpc_port = config.getint('TrackerRPC', 'Port')
    tracker_host = config.get('Tracker', 'Host')
    tracker_port = config.getint('Tracker', 'Port')
    tracker_config = config_section_as_dict(config, 'Tracker')
    admin_enabled = config.getint('RemoteAdminLogin', 'Enabled')
    admin_port = config.getint('RemoteAdminLogin', 'Port')
    database_url = config.get('Database', 'URL')

    # initalize database
    log.info('initializing database ...')
    init_db(database_url, echo=False)

    # initialize services
    log.info('initializing services ...')
    tracker = Tracker(tracker_config, (tracker_host, tracker_port),
                      webseed_port, key_file, cert_file)
    rpc = TrackerRPC((rpc_host, rpc_port), tracker, key_file, cert_file)
    webseed = Webseeder(webseed_port, tracker,
                        config.getint('Uploader', 'Max_handler_threads'),
                        opts.maxupload)
    admin = AdminRPC(('', admin_port), tracker, rpc, key_file, cert_file,
                     admin_enabled)

    # start threads
    log.info('creating TrackerRPC thread ...')
    rpc_thread = Thread(target=rpc.serve_forever, name='btdt tracker RPC')
    rpc_thread.start()

    log.info('creating Webseeder thread ...')
    webseed_thread = Thread(target=webseed.serve_forever,
                            name='btdt webseed')
    webseed_thread.start()

    if admin_enabled:
        log.info('creating AdminRPC thread ...')
        admin_thread = Thread(target=admin.serve_forever,
                              name='btdt admin RPC')
        admin_thread.start()

    log.info('btdt-server running')

    def print_(f):
        def impl(*args):
            res = f(*args)
            print res
            return res

        return impl

    def print_help(*args):
        h = (
            ('h(elp)', "prints this list"),
            ('q(quit)', "terminates the server"),
            ('adduser name password',
             "adds a new user or updates the user's password"),
            ('deleteruser name', "deletes an existing user"),
            ('listusers', "lists all users"),
            ('addadmin name', "gives admin rights to an existing user"),
            ('deleteadmin name', "removes admin rights from an existing user"),
            ('listadmins', "lists all admins"),
            ('u(pdate)', "updates the list of torrents")
        )

        for t in h:
            print '%s:\t%s' % t

    cmds = {
        'help': print_help,
        'quit': AdminRPC.shutdown_server,
        'adduser': AdminRPC.add_user,
        'deleteuser': AdminRPC.delete_user,
        'listusers': print_(AdminRPC.list_users),
        'addadmin': AdminRPC.add_admin,
        'deleteadmin': AdminRPC.delete_admin,
        'listadmins': print_(AdminRPC.list_admins),
        'update': AdminRPC.update_torrents
    }

    scmds = {
        'q': 'quit',
        'h': 'help',
        'u': 'update'
    }

    # handle user input
    try:
        while True:
            t = raw_input('> ')
            args = t.strip().split()
            if not args:
                continue

            cmd = args.pop(0)
            try:
                if not cmd:
                    continue
                elif cmd in cmds:
                    cmds[cmd](admin, *args)
                elif cmd in scmds:
                    cmds[scmds[cmd]](admin, *args)
                else:
                    log.error("command '%s' unknown (try help)" % cmd)
            except (KeyboardInterrupt, EOFError):
                raise
            except Exception, e:
                log.error("'%s' failed (exception follows):" % cmd)
                log.exception(e)
    except (KeyboardInterrupt, EOFError):
        pass

    # shutdown

    log.info('shutting down ...')
    if admin_enabled:
        admin.shutdown()
    webseed.shutdown()
    rpc.shutdown()
    tracker.shutdown()

    log.info('joining threads ...')
    if admin_enabled:
        log.debug('joining admin RPC thread ...')
        admin_thread.join()
    log.debug('joiuning webseeder thread ...')
    webseed_thread.join()
    log.debug('joining RPC thread ...')
    rpc_thread.join()

    log.info('===== bye bye =====')
    return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
