#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2009, Sebastian Ramacher
# Copyright (C) 2008, Andreas Stuehrk
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

""" libtorrent wrapper

Wraps some functions from libtorrent.
"""

from __future__ import with_statement

from btdt.common import size_to_human_readable, read_torrent
from btdt.config import fill_config_default_values
from btdt.logger import logger
from btdt.metafile import to_utf8
from tempfile import mktemp
from threading import Event, Thread
from urllib import quote as urlquote
from urlparse import urljoin

import libtorrent as lt
import os.path

STATUS_STRING = (
    'queued for checking',
    'checking files',
    'NOT USED',
    'downloading',
    'finished',
    'seeding',
    'allocation'
)

FORCE_ENC = 0
ENABLE_ENC = 1
DISABLE_ENC = 2
PLAIN_ENC = 0
RC4_ENC = 1
BOTH_ENC = 2

ENC_MODES = (FORCE_ENC, ENABLE_ENC, DISABLE_ENC)
ENC_LEVEL_MODES = (PLAIN_ENC, RC4_ENC, BOTH_ENC)


class SessionState(object):
    def __init__(self, status):
        self.status = status

    def __getattr__(self, name):
        return getattr(self.status, name)

    total_downloaded = property(lambda self: self.total_download)
    total_uploaded = property(lambda self: self.total_upload)
    download_rate_kbs = property(lambda self: self.download_rate / 1024.0)
    upload_rate_kbs = property(lambda self: self.upload_rate / 1024.0)


class TorrentState(object):
    def __init__(self, status, name):
        self.status = status
        self.name = name

    def __getattr__(self, name):
        return getattr(self.status, name)

    @property
    def eta(self):
        size = self.total_wanted
        done = self.total_wanted_done
        speed = int(self.download_rate)
        if (size - done) == 0 or speed == 0:
            eta = 0
        else:
            eta = (size - done) / speed
        return eta

    is_paused = property(lambda self: self.paused)
    progress_percent = property(lambda self: self.progress * 100)
    state_string = property(lambda self: STATUS_STRING[self.state])
    download_rate_kbs = property(lambda self: self.download_rate / 1024.0)
    upload_rate_kbs = property(lambda self: self.upload_rate / 1024.0)


class TorrentManager(object):
    """ Torrent manager

    Manages the torrents and some libtorrent related settings.
    """

    def use_upnp(self, val):
        if val:
            self.__session.start_upnp()
        else:
            self.__session.stop_upnp()

    def use_natpmp(self, val):
        if val:
            self.__session.start_natpmp()
        else:
            self.__session.stop_natpmp()

    def use_lsd(self, val):
        if val:
            self.__session.start_lsd()
        else:
            self.__session.stop_lsd()

    def __init__(self, ports,
                 pe_settings=(FORCE_ENC, FORCE_ENC, BOTH_ENC, True),
                 maxupload=0, use_upnp=False, use_lsd=False, use_natpmp=False,
                 fast_resume_path=None, webseed_workaround=0):
        """ Initializes libtorrent and sets some options """

        self.__session = lt.session(lt.fingerprint("BT", 0, 1, 0, 0))

        self.pe_settings(*pe_settings)
        self.listen(*ports)
        self.upload_rate_limit(maxupload)

        self.__session.stop_dht()
        self.use_upnp(use_upnp)
        self.use_lsd(use_lsd)
        self.use_natpmp(use_natpmp)

        self.fast_resume_path = None
        if fast_resume_path is not None and len(fast_resume_path):
            if not os.path.exists(fast_resume_path):
                os.mkdir(fast_resume_path)
            if os.path.isdir(fast_resume_path):
                self.fast_resume_path = fast_resume_path

        self.torrents = {}

        self.webseed_thread = None
        if webseed_workaround:
            self.webseed_workaround_interval = webseed_workaround * 60
            self.doneevent = Event();
            self.webseed_thread = Thread(target=self.__webseed_workaround,
                                         name='btdt DM webseed bug fix')
            self.webseed_thread.start()

    def quit(self, fastresume=True):
        """ Deinitializes deluge's core and saves fast resume data. """

        if self.webseed_thread is not None:
            self.doneevent.set()
            self.webseed_thread.join()

        # remove torrents
        for id in tuple(self.ids()):
            self.remove_torrent(id, fastresume=fastresume, delete_data=False)

        # self.__session.abort()
        # since abort doesn't seem to exist we just del it
        del self.__session

    def listen(self, minport, maxport=None):
        """
        Tells libtorrent to listen on one of the ports in [minport, maxport] or
        on minport if maxport is None.
        """

        if maxport is None:
            ports = [minport, minport]
        else:
            ports = sorted([minport, maxport])

        logger().debug("listening on %s" % str(ports))
        self.__session.listen_on(*ports)

    def listening_info(self):
        """
        Retuns False if Deluge is not listening on any port or the open
        port otherwise.
        """

        if self.__session.is_listening():
            return self.__session.listen_port()
        else:
            return False

    def pe_settings(self, encout=FORCE_ENC, encin=FORCE_ENC, level=BOTH_ENC,
                    prefer_rc4=True):
        if encout not in ENC_MODES or encin not in ENC_MODES or \
            level not in ENC_LEVEL_MODES:
                raise ValueError('invalid encryption settings')

        pe = lt.pe_settings()
        pe.out_enc_policy = lt.enc_policy(encout)
        pe.in_enc_policy = lt.enc_policy(encin)
        pe.allowed_enc_level = lt.enc_level(level)

        self.__session.set_pe_settings(pe)

    def upload_rate_limit(self, limit):
        """
        Sets the upload rate limit (in KB/s). The upload rate is unlimited
        if the given limit is smaller or equal to 0.
        """

        if limit <= 0:
            limit = -1
        else:
            limit = limit * 1024

        if limit != -1:
            logger().debug("setting upload rate limit to %s/s" % \
                                       size_to_human_readable(limit))
        else:
            logger().debug("setting upload rate limit to unlimited")
        self.__session.set_upload_rate_limit(limit)

    def add_torrent(self, path, target, storage=1):
        """ Adds a torrent to the download queue. """

        if not os.path.exists(path):
            raise ValueError('torrent path doest not exist')
        logger().debug("adding torrent '%s'" % path)

        if storage == 2:
            logger().warn("Using storage mode compact (=2) may lead to "
                          "segfaults!")

        info_hash_hex = read_torrent(path)[2]
        if self.fast_resume_path is not None:
            fsd = os.path.join(self.fast_resume_path,
                               '%s.fastresume' % info_hash_hex)
            if os.path.exists(fsd):
                logger().debug("found fast resume data for '%s'" % path)
                with open("%s.fastresume" % path, 'wb') as fo:
                    with open(fsd, 'rb') as fi:
                        fo.write(fi.read())


        with open(path, "rb") as f:
            ti = lt.torrent_info(lt.bdecode(f.read()))

        aps = {}
        aps['ti'] = ti
        aps['paused'] = False
        aps['auto_managed'] = False
        aps['duplicate_is_error'] = False
        aps['storage_mode'] = lt.storage_mode_t(storage)
        aps['save_path'] = target

        h = self.__session.add_torrent(aps)
        id = 0
        for id in xrange(len(self.torrents) + 1):
            if id not in self.torrents:
                break
        logger().debug("added torrent '%s' with id '%d'" % (path, id))

        self.torrents[id] = {
            'path': path,
            'target': target,
            'infohash': info_hash_hex, # h.infohash
            'handle': h
        }

        state = self.get_torrent_state(id)
        self.torrents[id]['name'] = state.name

        if self.webseed_thread is not None:
            data = read_torrent(path)[0]
            if 'url-list' in data and len(data['url-list']):
                self.torrents[id]['urls'] = data['url-list']

        return id

    def pause_torrent(self, id):
        """ Pauses the download of a torrent. """

        id = self.check_or_find(id)
        state = self.get_torrent_state(id)
        if not state.is_paused:
            self.torrents[id]['handle'].pause()

    def remove_torrent(self, id, fastresume=True, delete_data=False):
        """ Removes a torrent from the download queue. """

        id = self.check_or_find(id)
        if fastresume and self.fast_resume_path is not None:
            info_hash_hex = self.torrents[id]['infohash']
            # save_fastresume adds .fastresume
            fsd = os.path.join(self.fast_resume_path, info_hash_hex)
            fsd2 = '%s.fastresume' % fsd

            if not delete_data:
                path = self.torrents[id]['path']
                logger().debug("saving fastresume data for '%s'" % path)
                # dc.save_fastresume(id, fsd)

                # paranoid mode with open(fsd2, 'r') as f: data =
                             # lt.bdecode(f.read())
                #data['peers'] = {}
                # with open(fsd2, 'w') as f: f.write(lt.bencode(data))
            elif os.path.exists(fsd2):
                os.unlink(fsd2)

        self.__session.remove_torrent(self.torrents[id]['handle'],
                                      lt.options_t.delete_files if delete_data
                                      else lt.options_t.none)
        del self.torrents[id]

    def resume_torrent(self, id):
        """ Resumes the download of a torrent. """

        id = self.check_or_find(id)
        state = self.get_torrent_state(id)
        if state.is_paused:
            self.torrents[id]['handle'].resume()

    def get_peer_info(self, id_):
        """ Retrieves some information on a torrent's peers. """
        id_ = self.check_or_find(id_)
        return self.torrents[id_]['handle'].get_peer_info()

    def get_torrent_state(self, id):
        """ Retrieves some information on a torrent's state. """

        id = self.check_or_find(id)
        handle = self.torrents[id]['handle']
        return TorrentState(handle.status(), handle.name())

    def get_state(self):
        """ Retrives some information on the session's state. """

        return SessionState(self.__session.status())

    def ids(self):
        """ Retrives the torrent's IDs. """

        return self.torrents.iterkeys()

    def check_id(self, id):
        """ Checks if the id exists. """

        return id in self.torrents

    def find_id(self, torrent):
        """ Finds the unique id based on the torrent's name. """

        for tor in self.torrents.itervalues():
            if tor['name'] == torrent:
                return id

        return None

    def check_or_find(self, id):
        """
        Checks if the id exists or if there's a torrent with that name. Returns
        the id on success and raises InvalidUniqueIDError on failure.
        """

        if self.check_id(id):
            return id

        nid = self.find_id(id)
        if nid is not None:
            return nid

        raise InvalidUniqueIDError("torrent with name or id '%s' doesn't exist"
                                   % str(id))

    def replace_trackers(self, id, trackers):
        """ Replaces the tracker of a specific torrent. """

        if isinstance(trackers, basestring):
            trackers = [trackers]

        ae = [lt.announce_entry(url) for url in trackers]

        id = self.check_or_find(id)
        return self.torrents[id]['handle'].replace_trackers(ae)

    def __webseed_workaround(self):
        while not self.doneevent.isSet():
            for id, info in self.torrents.iteritems():
                if 'urls' not in info:
                    continue

                state = self.get_torrent_state(id)
                if state.is_paused or state.state_string != 'downloading':
                    continue

                webseeds = self.torrents[id]['handle'].url_seeds()
                urls = info['urls']
                if isinstance(urls, basestring):
                    urls = [urls]

                logger().debug("webseeds are '%s', should be '%s'" % \
                               (str(webseeds), str(urls)))
                for url in urls:
                    found = False
                    for webseed in webseeds:
                        if webseed in url:
                            found = True

                    if not found:
                        logger().info("readding webseed '%s'" % url)
                        self.torrents[id]['handle'].add_url_seed(url)

            self.doneevent.wait(self.webseed_workaround_interval)

from metafile import make_meta_file

def create_torrent(destination, input, tracker=None, comment=None,
                   piece_size=1024, creator='BTDT', priv=False, webseeds=None):
    """
    Creates a torrent and takes the same arguments as deluge.deluge_core's
    create_torrent.
    """

    if not os.path.exists(input):
        raise ValueError("input path doesn't exist")

    # since libtorrent changed it's web seed behaviour, we're fixing the
    # webseed urls for single file torrents
    def fix_url(url, filename):
        if not url.endswith(filename):
            return urljoin(url, urlquote(to_utf8(filename)))
        else:
            return url

    if not os.path.isdir(input):
        filename = os.path.basename(input)
        if isinstance(webseeds, basestring):
            webseeds = fix_url(webseeds, filename)
        else:
            try:
                webseeds = [fix_url(url, filename) for url in webseeds]
            except TypeError:
                pass

    dest = mktemp()
    make_meta_file(input, tracker, piece_size, target=dest, comment=comment,
                   url_list=webseeds, created_by=creator, private=priv,
                   httpseeds=webseeds)

    with open(destination, 'wb') as fo:
        with open(dest, 'rb') as fi:
            fo.write(fi.read())
