#!/usr/bin/env python
# -*- coding: utf-8 -*-
# $Id$
# Copyright (C) 2008, Sebastian Ramacher
# Copyright (C) 2008, Andreas Stuehrk
# This file is released under a modified BSD license. See LICENSE.txt for
# details.
################################################################################

"""
local proxy which forwards requests to a tracker via secured XMLRPC
"""

import socket

from threading import Event
from BaseHTTPServer import HTTPServer
from httplib import HTTPConnection
from SimpleHTTPServer import SimpleHTTPRequestHandler
from urllib import unquote
from xmlrpclib import Binary

from btdt.common import AuthorizedServerProxy, furlparse

class TrackerProxyRequestHandler(SimpleHTTPRequestHandler):
    protocol_version = 'HTTP/1.1'

    get_handlers = list()

    def log_request(self, *args, **kwargs):
       pass

    def register_handler(path_matcher, handlers=get_handlers):
        if isinstance(path_matcher, basestring):
            path = path_matcher
            path_matcher = lambda path, match_path=path: path == match_path
        def decorator(func):
            handlers.append((path_matcher, func))
            return func
        return decorator

    def send_headers(self, headers, end_headers=False):
        for key, value in headers.iteritems():
            self.send_header(key, value)
        if end_headers:
            self.end_headers()

    def send(self, code, headers, data):
        try:
            if not 'Content-Length' in headers:
                headers['Content-Length'] = len(data)

            self.send_response(code)
            self.send_headers(headers, True)
            self.wfile.write(data)
        except socket.error:
            pass

    def do_QUIT(self):
        # just handle the request
        self.server.doneevent.set()
        return self.send(200, {}, '')

    def do_GET(self):
        try:
            (scheme, netloc, path, pars, query, fragment) = furlparse(self.path)
            path = unquote(path)[1:]
            params = {}
            for s in query.split('&'):
                if s != '':
                    i = s.index('=')
                    params[unquote(s[:i])] = unquote(s[i+1:])

            for path_matcher, handler in self.get_handlers:
                if path_matcher(path):
                    return handler(self, path, params)

            return self.send_error(404, 'File not found')

        except socket.error, e:
            self.server.reconnect()
            return self.send_error(500, 'socket error - ' + str(e))
        except ValueError, e:
            return self.send_error(400, 'you sent me garbage - ' + str(e))
        except Exception, e:
            return self.send_error(500, 'some unhandled error - ' + str(e))

    @register_handler('announce')
    def handle_announce(self, path, params):
        try:
            infohash = params['info_hash']
            del params['info_hash']
        except KeyError:
            raise ValueError('no info hash')

        code, headers, data = self.server.rpc.announce(Binary(infohash), params)
        return self.send(code, headers, data.data)

    @register_handler('scrape')
    def handle_scrape(self, path, params):
        infohash = None
        if 'info_hash' in params:
            infohash = params['info_hash']
            del params['info_hash']

        code, headers, data = self.server.rpc.scrape(Binary(infohash), params)
        return self.send(code, headers, data.data)

    del register_handler

class TrackerProxy(HTTPServer):
    def __init__(self, port, xmlrpc_url, username, password):
        self.__connect_func = lambda: AuthorizedServerProxy(xmlrpc_url,
                                                            username, password)
        self.reconnect()
        self.doneevent = Event()
        HTTPServer.__init__(self, ('localhost', port),
                            TrackerProxyRequestHandler)

    def reconnect(self):
        self.rpc = self.__connect_func()

    def server_bind(self):
        if self.allow_reuse_address:
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if self.server_address[1] < 0:
            self.server_address = (self.server_address[0], socket.INADDR_ANY)
        self.socket.bind(self.server_address)
        self.server_address = self.socket.getsockname()
        host, port = self.socket.getsockname()[:2]
        self.server_name = socket.getfqdn(host)
        self.server_port = port

    def serve_forever(self):
        while not self.doneevent.isSet():
            self.handle_request()

    def shutdown(self):
        # just force the server to handle a request
        conn = HTTPConnection('localhost:%d' % self.server_port)
        conn.request("QUIT", "/")
        conn.getresponse()
